import numpy as np

offset_amp = 2**-3  # (0.125)


def _iq_imbalance_corr(g, phi):
    c = np.cos(phi)
    s = np.sin(phi)
    n = 1 / ((1 - g**2) * (2 * c**2 - 1))
    return [float(n * x) for x in [(1 - g) * c, (1 + g) * s, (1 - g) * s, (1 + g) * c]]


def _get_frequencies(if_freq, optimizer_parameters):
    down_mixer_offset = optimizer_parameters["calibration_offset_frequency"]
    signal_freq = if_freq - down_mixer_offset
    image_freq = -if_freq - down_mixer_offset
    return down_mixer_offset, signal_freq, image_freq


def _prep_config(iq_channels, adc_channels, if_freq, lo_freq, optimizer_parameters):
    i0_offset = 0
    q0_offset = 0
    g = 0.0
    phi = 0.0
    calibration_amp = 0.125
    calibration_pulse_length = 10e3
    time_of_flight = 192
    correction_matrix = _iq_imbalance_corr(g, phi)

    down_mixer_offset, signal_freq, image_freq = _get_frequencies(if_freq, optimizer_parameters)

    i_port = iq_channels[0]
    q_port = iq_channels[1]
    controller_names = set([ch[0] for ch in iq_channels] + [ch[0] for ch in adc_channels])
    controllers = {
        name: {
            "type": "opx1",
        }
        for name in controller_names
    }
    controllers[i_port[0]]["analog_outputs"] = {}
    controllers[i_port[0]]["analog_outputs"][i_port[1]] = {"offset": i0_offset}
    controllers[q_port[0]]["analog_outputs"][q_port[1]] = {"offset": q0_offset}
    for port in adc_channels:
        controllers[port[0]]["analog_inputs"] = {}
    for port in adc_channels:
        controllers[port[0]]["analog_inputs"][port[1]] = {"offset": 0.0}

    config = {
        "version": 1,
        "controllers": controllers,
        "elements": {
            "IQmixer": {  # IQ mixer element
                "mixInputs": {
                    "I": i_port,
                    "Q": q_port,
                    "lo_frequency": lo_freq,
                    "mixer": "Correction_mixer",
                },
                "intermediate_frequency": if_freq,
                "operations": {
                    "calibration": "calibration_pulse",
                    "calibration_long": "long_calibration_pulse",
                },
                "digitalInputs": {},  # This is being added afterwards
            },
            "I_offset": {  # element to apply the DC offset on I
                "singleInput": {"port": i_port},
                "hold_offset": {"duration": 1},
                "operations": {"DC_offset": "DC_offset_pulse"},
            },
            "Q_offset": {  # element to apply the DC offset on Q
                "singleInput": {"port": q_port},
                "hold_offset": {"duration": 1},
                "operations": {"DC_offset": "DC_offset_pulse"},
            },
            "signal_analyzer": {
                "intermediate_frequency": signal_freq,
                "mixInputs": {
                    "I": i_port,
                    "Q": q_port,
                    "lo_frequency": lo_freq,
                    "mixer": "Correction_mixer",
                },
                "operations": {
                    "Analyze": "Analyze_pulse",
                },
                "outputs": {"out1": adc_channels[0], "out2": adc_channels[1]},
                "time_of_flight": time_of_flight,
                "smearing": 0,
            },
            "lo_analyzer": {
                "intermediate_frequency": -down_mixer_offset,
                "mixInputs": {
                    "I": i_port,
                    "Q": q_port,
                    "lo_frequency": lo_freq,
                    "mixer": "Correction_mixer",
                },
                "operations": {
                    "Analyze": "Analyze_pulse",
                },
                "outputs": {"out1": adc_channels[0], "out2": adc_channels[1]},
                "time_of_flight": time_of_flight,
                "smearing": 0,
            },
            "image_analyzer": {
                "intermediate_frequency": image_freq,
                "mixInputs": {
                    "I": i_port,
                    "Q": q_port,
                    "lo_frequency": lo_freq,
                    "mixer": "Correction_mixer",
                },
                "operations": {
                    "Analyze": "Analyze_pulse",
                },
                "outputs": {"out1": adc_channels[0], "out2": adc_channels[1]},
                "time_of_flight": time_of_flight,
                "smearing": 0,
            },
        },
        "pulses": {
            "calibration_pulse": {
                "operation": "control",
                "length": calibration_pulse_length,
                "waveforms": {
                    "I": "readout_wf",
                    "Q": "zero_wf",
                },
                "digital_marker": "ON",
            },
            "long_calibration_pulse": {
                "operation": "control",
                "length": calibration_pulse_length * 100,
                "waveforms": {
                    "I": "readout_wf",
                    "Q": "zero_wf",
                },
                "digital_marker": "ON",
            },
            "DC_offset_pulse": {
                "operation": "control",
                "length": calibration_pulse_length,
                "waveforms": {"single": "DC_offset_wf"},
            },
            "Analyze_pulse": {
                "operation": "measurement",
                "length": calibration_pulse_length,
                "waveforms": {
                    "I": "zero_wf",
                    "Q": "zero_wf",
                },
                "integration_weights": {
                    "integW_cos": "integW_cosine",
                    "integW_sin": "integW_sine",
                    "integW_minus_sin": "integW_minus_sine",
                },
                "digital_marker": "ON",
            },
        },
        "waveforms": {
            "readout_wf": {
                "type": "constant",
                "sample": calibration_amp,
            },
            "zero_wf": {
                "type": "constant",
                "sample": 0.0,
            },
            "DC_offset_wf": {"type": "constant", "sample": offset_amp},
        },
        "digital_waveforms": {
            "ON": {"samples": [(1, 0)]},
            "OFF": {"samples": [(0, 0)]},
        },
        "integration_weights": {
            "integW_cosine": {
                "cosine": [1.0] * int(calibration_pulse_length / 4),
                "sine": [0.0] * int(calibration_pulse_length / 4),
            },
            "integW_sine": {
                "cosine": [0.0] * int(calibration_pulse_length / 4),
                "sine": [1.0] * int(calibration_pulse_length / 4),
            },
            "integW_minus_sine": {
                "cosine": [0.0] * int(calibration_pulse_length / 4),
                "sine": [-1.0] * int(calibration_pulse_length / 4),
            },
        },
        "mixers": {
            "Correction_mixer": [
                {
                    "intermediate_frequency": if_freq,
                    "lo_frequency": lo_freq,
                    "correction": correction_matrix,
                },
                {
                    "intermediate_frequency": signal_freq,
                    "lo_frequency": lo_freq,
                    "correction": correction_matrix,
                },
                {
                    "intermediate_frequency": -down_mixer_offset,
                    "lo_frequency": lo_freq,
                    "correction": correction_matrix,
                },
                {
                    "intermediate_frequency": image_freq,
                    "lo_frequency": lo_freq,
                    "correction": correction_matrix,
                },
            ],
        },
    }

    return config



import pathlib
import numpy
from typing import TypeVar, Union, ClassVar
from typing_extensions import Protocol

PathLike = TypeVar("PathLike", str, pathlib.Path)
Number = Union[int, float]
Value = Union[Number, bool]

NumpyNumber = Union[numpy.floating, numpy.integer]
NumpyValue = Union[NumpyNumber, numpy.bool_]

NumpySupportedNumber = Union[Number, NumpyNumber]
NumpySupportedFloat = Union[float, numpy.floating]
NumpySupportedValue = Union[Value, NumpyValue]


class DataClassType(Protocol):
    __dataclass_fields__: ClassVar[dict]



from qm.grpc.frontend import (
    SimulationRequest,
    ExecutionRequestSimulateSimulationInterfaceQuantumSimulatorInterface,
)
from qm.simulate.interface import SimulatorInterface
from qm.grpc.quantum_simulator.v1 import PhysicalConfig


class QuantumSimulatorInterface(SimulatorInterface):
    def __init__(self, physical_config: PhysicalConfig) -> None:
        super().__init__()
        self._physical_config = physical_config

    def update_simulate_request(self, request: SimulationRequest) -> SimulationRequest:
        request = super().update_simulate_request(request)
        request.simulate.simulation_interface.quantum_simulator = (
            ExecutionRequestSimulateSimulationInterfaceQuantumSimulatorInterface(
                physical_config=bytes(self._physical_config)
            )
        )
        return request



from typing import Tuple, Union, List, Optional, Sequence

import betterproto
import numpy
from dependency_injector.wiring import Provide, inject

from qm.api.frontend_api import FrontendApi
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.devices import AnalogOutputPortFilter, MixerInfo
from qm.containers.capabilities_container import CapabilitiesContainer
from qm.elements.basic_element import logger, BasicElement
from qm.grpc.qua_config import QuaConfigDacPortReference
from qm.type_hinting.general import NumpySupportedFloat
from qm.utils import fix_object_data_type
from qm.grpc.general_messages import Matrix


def _fix_offset(offset):
    if offset == 0:
        offset = float(offset)
    if not isinstance(offset, (numpy.floating, float)):
        raise TypeError("offset must be a float or a tuple of floats")

    offset = fix_object_data_type(offset)
    return offset


def _set_single_output_port_dc_offset(
    frontend_api: FrontendApi,
    machine_id: str,
    element_name: str,
    input_name: str,
    offset,
):
    offset = _fix_offset(offset)
    logger.debug(f"Setting DC offset of input '{input_name}' on element '{element_name}' to '{offset}'")
    frontend_api.set_output_dc_offset(machine_id, element_name, input_name, offset)


def _create_taps_filter(feedforward, feedback):
    for name, instance in zip(["feedforward", "feedback"], [feedforward, feedback]):
        if instance is not None and not isinstance(instance, (numpy.ndarray, List)):
            raise TypeError(f"{name} must be None, a list, or a numpy array. Got {type(instance)}.")
    if isinstance(feedforward, numpy.ndarray):
        feedforward = feedforward.tolist()
    if isinstance(feedback, numpy.ndarray):
        feedback = feedback.tolist()
    return AnalogOutputPortFilter(feedforward=feedforward, feedback=feedback)


@inject
def static_set_mixer_correction(
    frontend_api: FrontendApi,
    machine_id: str,
    mixer: str,
    intermediate_frequency: Union[int, float],
    lo_frequency: Union[int, float],
    values: Tuple[float, float, float, float],
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
):
    # TODO - this function is here (and not under MixedInputsElement) to support backwards the direct calling to mixer
    #  Once it is changed, one can put this function under the element

    if not isinstance(values, (tuple, list)) or len(values) != 4:
        raise Exception("correction values must have 4 items")

    correction_matrix = Matrix(*[fix_object_data_type(x) for x in values])

    if capabilities.supports_double_frequency:
        frequencies = {
            "lo_frequency_double": float(lo_frequency),
            "intermediate_frequency_double": abs(float(intermediate_frequency)),
        }
    else:
        frequencies = {
            "lo_frequency": int(lo_frequency),
            "intermediate_frequency": abs(int(intermediate_frequency)),
        }

    mixer_info = MixerInfo(
        mixer=mixer,
        frequency_negative=intermediate_frequency < 0,
        **frequencies,
    )
    frontend_api.set_correction(machine_id, mixer_info, correction_matrix)


class SingleInputElement(BasicElement):
    @property
    def port(self) -> QuaConfigDacPortReference:
        return self._config.single_input.port

    def get_output_dc_offset(self) -> float:
        pass

    def set_output_dc_offset(self, offset: float):
        _set_single_output_port_dc_offset(self._frontend, self._id, self._name, "single", offset)

    def set_output_filter(
        self,
        feedforward: Union[Sequence[NumpySupportedFloat], None],
        feedback: Union[Sequence[NumpySupportedFloat], None],
    ):
        analog_filter = _create_taps_filter(feedforward, feedback)
        self._frontend.set_output_filter_taps(self._id, self._name, "single", analog_filter)


class MultipleInputsElement(BasicElement):
    pass


class SingleInputCollectionElement(BasicElement):
    pass


class MixInputsElement(BasicElement):
    @property
    def i_port(self) -> QuaConfigDacPortReference:
        return self._config.mix_inputs.i

    @property
    def q_port(self) -> QuaConfigDacPortReference:
        return self._config.mix_inputs.q

    @property
    def lo_frequency(self) -> int:
        return self._config.mix_inputs.lo_frequency

    @lo_frequency.setter
    @inject
    def lo_frequency(
        self, value: float, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
    ):
        freq = fix_object_data_type(value)
        logger.debug(f"Setting element '{self._name}' intermediate frequency to '{freq}'.")
        if capabilities.supports_double_frequency:
            self._config.mix_inputs.lo_frequency_double = float(freq)
            self._config.mix_inputs.lo_frequency = 0
        else:
            self._config.mix_inputs.lo_frequency = int(freq)
            self._config.mix_inputs.lo_frequency_double = 0.0

    @property
    def mixer(self) -> str:
        return self._config.mix_inputs.mixer

    @property
    def has_octave_params(self) -> bool:
        return betterproto.serialized_on_wire(self._config.mix_inputs.octave_params)

    def get_output_dc_offset(self, port_name) -> float:
        pass

    def set_output_dc_offset(self, i_offset: Optional[float] = None, q_offset: Optional[float] = None):
        if i_offset is not None:
            _set_single_output_port_dc_offset(self._frontend, self._id, self._name, "I", i_offset)
        if q_offset is not None:
            _set_single_output_port_dc_offset(self._frontend, self._id, self._name, "Q", q_offset)

    def set_output_filter(
        self,
        input_name: str,
        feedforward: Union[Sequence[NumpySupportedFloat], None],
        feedback: Union[Sequence[NumpySupportedFloat], None],
    ):
        analog_filter = _create_taps_filter(feedforward, feedback)
        self._frontend.set_output_filter_taps(
            self._id,
            self._name,
            input_name,
            analog_filter,
        )

    def set_mixer_correction(
        self,
        intermediate_frequency: int,
        lo_frequency: int,
        values: Tuple[float, float, float, float],
    ):
        static_set_mixer_correction(
            self._frontend,
            self._id,
            mixer=self.mixer,
            intermediate_frequency=intermediate_frequency,
            lo_frequency=lo_frequency,
            values=values,
        )



from qm.type_hinting.general import Number, Value, NumpySupportedValue, NumpySupportedNumber
from qm.type_hinting.config_types import DictQuaConfig


__all__ = ["DictQuaConfig", "Number", "Value", "NumpySupportedValue", "NumpySupportedNumber"]



import time
from typing import TypeVar, Callable, Optional

T = TypeVar("T")


def run_until_with_timeout(
    on_iteration_callback: Callable[[], bool],
    on_complete_callback: Callable[[], Optional[T]] = lambda: None,
    timeout: float = float("infinity"),
    loop_interval: float = 0.1,
    timeout_message: str = "Timeout Exceeded",
) -> Optional[T]:
    """

    :param on_iteration_callback: A callback that returns bool that is called every loop iteration.
     If True is returned, the loop is complete and on_complete_callback is called.

    :param on_complete_callback: A callback that is called when the loop is completed.
    This function returns the return value of on_complete_callback

    :param timeout: The timeout in seconds for on_iteration_callback to return True,
    on Zero the loop is executed once.

    :param loop_interval: The interval (in seconds) between each loop execution.

    :param timeout_message: The message of the TimeoutError exception.
    raise TimeoutError: When the timeout is exceeded
    """
    if timeout < 0:
        raise ValueError("timeout cannot be smaller than 0")

    start = time.time()
    end = start + timeout

    while True:
        if on_iteration_callback():
            return on_complete_callback()

        time.sleep(loop_interval)

        if time.time() >= end:
            raise TimeoutError(timeout_message)



from typing import Dict, Union, Optional

import betterproto
from octave_sdk import OctaveLOSource, OctaveOutput, Octave

from qm.api.frontend_api import FrontendApi
from qm.elements.basic_element import BasicElement
from qm.elements.native_elements import (
    SingleInputElement,
    MixInputsElement,
    MultipleInputsElement,
    SingleInputCollectionElement,
)
from qm.elements.element_with_octave import ElementWithOctave
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigMultipleInputs,
    QuaConfigMixInputs,
    QuaConfigSingleInput,
    QuaConfigSingleInputCollection,
    QuaConfigDacPortReference,
)
from qm.grpc.octave_models import OctaveIfInputPort, OctaveIfInputPortName
from qm.octave import QmOctaveConfig


MaybeOctaveInputPort = Optional[OctaveIfInputPort]


class ElementNotFound(KeyError):
    def __init__(self, key):
        self._key = key

    def __str__(self) -> str:
        return f"Element with the key {self._key} was not found."


class UnknownElementType(ValueError):
    pass


class OctaveLoopbackError(Exception):
    pass


class OctaveConnectionError(Exception):
    def __init__(self, i_port: MaybeOctaveInputPort, q_port: MaybeOctaveInputPort):
        self._i = i_port
        self._q = q_port


class OnlyOneConnectedPort(OctaveConnectionError):
    def __str__(self) -> str:
        connected_port = self._i or self._q
        return (
            f"Only {connected_port.port_name.name} is connected to octave, while the other isn't. "
            f"Either both of them should be connected, or none of them"
        )


class DifferentOctavesForTheSameElement(OctaveConnectionError):
    def __str__(self) -> str:
        return (
            f"The I port is connected to {self._i.device_name} and Q is connected to {self._q.device_name}. "
            f"Both of them should be connected to the same Octave, with the same number (i.e. I1 & Q1 etc.)."
        )


class DifferentOctaveInputsForTheSameElement(OctaveConnectionError):
    def __str__(self) -> str:
        return (
            f"The I port is connected to {self._i.port_name.name} and Q is connected to {self._q.port_name.name}. "
            f"Both of them should be connected to the same Octave, with the same number (i.e. I1 & Q1 etc.)."
        )


class BothPortsConnectedToTheSameOctavePort(OctaveConnectionError):
    def __str__(self) -> str:
        return (
            f"Both I and Q port are connected to {(self._i.device_name, self._i.port_name.name)}. "
            f"This is probably a configuration mistake."
        )


class ElementsDB(Dict[str, BasicElement]):
    def __missing__(self, key) -> None:
        raise ElementNotFound(key)


INPUTS_TO_ELEMENTS = {
    type(None): BasicElement,
    QuaConfigMixInputs: MixInputsElement,
    QuaConfigSingleInput: SingleInputElement,
    QuaConfigMultipleInputs: MultipleInputsElement,
    QuaConfigSingleInputCollection: SingleInputCollectionElement,
}


def init_elements(
    pb_config: QuaConfig,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_config: Optional[QmOctaveConfig] = None,
) -> ElementsDB:
    elements = {}
    _octave_manager = _OctavesContainer(pb_config, octave_config)
    for name, element_config in pb_config.v1_beta.elements.items():
        _, element_inputs = betterproto.which_one_of(element_config, "element_inputs_one_of")
        try:
            element_class = INPUTS_TO_ELEMENTS[type(element_inputs)]
        except KeyError:
            raise UnknownElementType(f"Element {name} is of unknown type - {type(element_inputs)}.")
        element = element_class(name, element_config, frontend_api, machine_id)
        element = _octave_manager.add_octave(element)

        elements[name] = element
    return ElementsDB(elements)


class _OctavesContainer:
    def __init__(self, pb_config: QuaConfig, octave_config: Optional[QmOctaveConfig] = None):
        self._pb_config = pb_config
        self._octave_config = octave_config or QmOctaveConfig()
        self._octave_clients_cache: Dict[str, Octave] = {}

    def _get_connections_to_octave(self, element_port: QuaConfigDacPortReference) -> MaybeOctaveInputPort:
        controller, number = element_port.controller, element_port.number
        controller_config = self._pb_config.v1_beta.controllers[controller]
        connection = controller_config.analog_outputs[number].octave_connectivity
        if betterproto.serialized_on_wire(connection):
            return connection
        res = self._octave_config.get_octave_input_port(controller, number)
        if res is None:
            return None
        octave_name, port_name = res
        return OctaveIfInputPort(device_name=octave_name, port_name=OctaveIfInputPortName[port_name])

    def _get_loopbacks(self, octave_name: str) -> Dict[OctaveLOSource, OctaveOutput]:
        if octave_name in self._pb_config.v1_beta.octaves:
            pb_loopbacks = self._pb_config.v1_beta.octaves[octave_name].loopbacks
            if pb_loopbacks:
                loopbacks = {}
                for loopback in pb_loopbacks:
                    source_octave = loopback.lo_source_generator.device_name
                    if source_octave != octave_name:
                        raise OctaveLoopbackError("lo loopback between different octave devices are not supported")
                    lo_source_input = OctaveLOSource[loopback.lo_source_input.name]
                    loopbacks[lo_source_input] = OctaveOutput[loopback.lo_source_generator.port_name.name]
                return loopbacks
            else:
                return {}
        return self._octave_config.get_lo_loopbacks_by_octave(octave_name)

    @staticmethod
    def _validate_octave_is_connected_properly(
        i_octave_input_port: MaybeOctaveInputPort, q_octave_input_port: MaybeOctaveInputPort
    ):
        if i_octave_input_port is None and q_octave_input_port is None:
            return
        if (i_octave_input_port is None) != (q_octave_input_port is None):
            raise OnlyOneConnectedPort(i_octave_input_port, q_octave_input_port)
        if i_octave_input_port.device_name != q_octave_input_port.device_name:
            raise DifferentOctavesForTheSameElement(i_octave_input_port, q_octave_input_port)
        if i_octave_input_port.port_name.name[-1] != q_octave_input_port.port_name.name[-1]:
            raise DifferentOctaveInputsForTheSameElement(i_octave_input_port, q_octave_input_port)
        if i_octave_input_port == q_octave_input_port:
            raise BothPortsConnectedToTheSameOctavePort(i_octave_input_port, q_octave_input_port)

    def add_octave(
        self,
        element: Union[MixInputsElement, SingleInputElement, MultipleInputsElement, SingleInputCollectionElement],
    ) -> Union[
        MixInputsElement,
        SingleInputElement,
        MultipleInputsElement,
        SingleInputCollectionElement,
        ElementWithOctave,
    ]:
        if not isinstance(element, MixInputsElement):
            return element

        i_octave_input_port = self._get_connections_to_octave(element.i_port)
        q_octave_input_port = self._get_connections_to_octave(element.q_port)

        self._validate_octave_is_connected_properly(i_octave_input_port, q_octave_input_port)
        if i_octave_input_port is None:
            if element.has_octave_params:
                raise ValueError("Element has octave params even though it is not connected to an Octave.")
            return element

        # TODO - add this when we fully move to configuration of the octave in the qua-config
        # if not element.has_octave_params:
        #     raise ValueError("Element does not have octave params even though it is connected to an Octave.")
        client = self._get_octave_client(i_octave_input_port.device_name)
        if betterproto.serialized_on_wire(element._config.mix_inputs.octave_params.rf_input_port):
            downconversion_client_name = element._config.mix_inputs.octave_params.rf_input_port.device_name
            downconversion_client = self._get_octave_client(downconversion_client_name)
            downconversion_port_number = element._config.mix_inputs.octave_params.rf_input_port.port_name.value + 1
        else:
            downconversion_client, downconversion_port_number = None, None
        return ElementWithOctave(
            element._name,
            element._config,
            element._frontend,
            element._id,
            client=client,
            octave_if_input_port_number=int(i_octave_input_port.port_name.name[-1]),
            downconversion_client=downconversion_client,
            octave_rf_input_port_number=downconversion_port_number,
        )

    def _get_octave_client(self, device_name: str) -> Octave:
        if device_name in self._octave_clients_cache:
            return self._octave_clients_cache[device_name]

        device_connection_info = self._octave_config.devices[device_name]
        loopbacks = self._get_loopbacks(device_name)
        client = Octave(
            host=device_connection_info.host,
            port=device_connection_info.port,
            port_mapping=loopbacks,
            octave_name=device_name,
            fan=self._octave_config.fan,
        )
        self._octave_clients_cache[device_name] = client
        return client



from qm.utils.async_utils import run_async
from qm.api.base_api import BaseApi, connection_error_handle
from qm.api.models.info import ImplementationInfo, QuaMachineInfo
from qm.api.models.server_details import ConnectionDetails
from qm.io.qualang.api.v1 import InfoServiceStub, GetInfoRequest


@connection_error_handle()
class InfoServiceApi(BaseApi):
    def __init__(self, connection_details: ConnectionDetails):
        super().__init__(connection_details)
        self._stub = InfoServiceStub(self._channel)

    def get_info(self) -> QuaMachineInfo:
        request = GetInfoRequest()
        response = run_async(self._stub.get_info(request, timeout=self._timeout))

        return QuaMachineInfo(
            capabilities=response.capabilities,
            implementation=ImplementationInfo(
                name=response.implementation.name,
                version=response.implementation.version,
                url=response.implementation.url,
            ),
        )



from dataclasses import dataclass

from qm.grpc import qm_manager


@dataclass
class Controller:
    name: str

    @staticmethod
    def build_from_message(message: qm_manager.Controller) -> "Controller":
        return Controller(message.name)



from marshmallow import ValidationError
from qm.exceptions import ConfigValidationException


def validate_timetagging_parameters(data):
    if "outputPulseParameters" in data:
        pulseParameters = data["outputPulseParameters"]
        neededParameters = [
            "signalThreshold",
            "signalPolarity",
            "derivativeThreshold",
            "derivativePolarity",
        ]
        missingParameters = []
        for parameter in neededParameters:
            if parameter not in pulseParameters:
                missingParameters.append(parameter)
        if len(missingParameters) > 0:
            raise ConfigValidationException(
                "An element defining the output pulse parameters must either "
                f"define all of the parameters: {neededParameters}. "
                f"Parameters defined: {pulseParameters}"
            )
        validPolarity = {"ASCENDING", "DESCENDING", "ABOVE", "BELOW"}
        if data["outputPulseParameters"]["signalPolarity"].upper() not in validPolarity:
            raise ConfigValidationException(
                f"'signalPolarity' is {data['outputPulseParameters']['signalPolarity'].upper()} but it must be one of {validPolarity}"
            )
        if data["outputPulseParameters"]["derivativePolarity"].upper() not in validPolarity:
            raise ConfigValidationException(
                f"'derivativePolarity' is {data['outputPulseParameters']['derivativePolarity'].upper()} but it must be one of {validPolarity}"
            )


def validate_output_tof(data):
    if "outputs" in data and data["outputs"] != {} and "time_of_flight" not in data:
        raise ValidationError("An element with an output must have time_of_flight defined")
    if "outputs" not in data and "time_of_flight" in data:
        raise ValidationError("time_of_flight should be used only with elements that have outputs")


def validate_output_smearing(data):
    if "outputs" in data and data["outputs"] != {} and "smearing" not in data:
        raise ValidationError("An element with an output must have smearing defined")
    if "outputs" not in data and "smearing" in data:
        raise ValidationError("smearing should be used only with elements that have outputs")


def validate_oscillator(data):
    if "intermediate_frequency" in data and "oscillator" in data:
        raise ValidationError("'intermediate_frequency' and 'oscillator' cannot be defined together")


def validate_used_inputs(data):
    used_inputs = list(
        filter(
            lambda it: it in data,
            ["singleInput", "mixInputs", "singleInputCollection", "multipleInputs"],
        )
    )
    if len(used_inputs) > 1:
        raise ValidationError(
            f"Can't support more than a single input type. " f"Used {', '.join(used_inputs)}",
            field_name="",
        )


def validate_arbitrary_waveform(is_overridable, has_max_allowed_error, has_sampling_rate):
    if is_overridable and has_max_allowed_error:
        raise ValidationError("Overridable waveforms cannot have property 'max_allowed_error'")
    if is_overridable and has_sampling_rate:
        raise ValidationError("Overridable waveforms cannot have property 'sampling_rate_key'")
    if has_max_allowed_error and has_sampling_rate:
        raise ValidationError("Cannot use both 'max_allowed_error' and 'sampling_rate'")



from typing import Optional

import betterproto
from octave_sdk import (
    OctaveLOSource,
    RFOutputMode,
    RFInputLOSource,
    IFMode,
    ClockType,
    ClockFrequency,
    RFInputRFSource,
)
from octave_sdk.octave import ClockInfo, Octave

from qm.api.frontend_api import FrontendApi
from qm.elements.native_elements import MixInputsElement
from qm.grpc.qua_config import QuaConfigElementDec


class SetFrequencyException(Exception):
    pass


class ElementWithOctave(MixInputsElement):
    def __init__(
        self,
        name: str,
        config: QuaConfigElementDec,
        frontend_api: FrontendApi,
        machine_id: str,
        client: Octave,
        octave_if_input_port_number: int,
        downconversion_client: Optional[Octave],
        octave_rf_input_port_number: Optional[int],
    ):
        super().__init__(name, config, frontend_api, machine_id)
        self._client = client
        self._octave_if_input_port_number_backup = octave_if_input_port_number
        self._downconversion_client = downconversion_client
        self._octave_rf_input_port_number = octave_rf_input_port_number

    def set_client(self) -> None:
        """
        This function is intended to be used from outside, using the batch mode,
        to run it in parallel on all the quantum-elements, on all the octaves
        """
        if self._config.mix_inputs.octave_params.lo_source:
            self.set_lo_source(OctaveLOSource[self._config.config.mix_inputs.octave_params.lo_source.name])
        else:
            self.set_lo_source(OctaveLOSource.Internal)

        if self.lo_source == OctaveLOSource.Internal and self.lo_source in self._client._port_mapping:
            self.set_lo_frequency(self.lo_frequency, set_source=False)

    def set_downconversion_port(self, downconversion_client: Octave, octave_rf_input_port_number: int):
        self._downconversion_client = downconversion_client
        self._octave_rf_input_port_number = octave_rf_input_port_number

    @property
    def _octave_if_input_port_number(self) -> int:
        if betterproto.serialized_on_wire(self._config.up_converted):
            enum_number = self._config.up_converted.rf_output_port.output.value + 1
            # The enums are 0-based and the ports are 1-based
        else:
            enum_number = self._octave_if_input_port_number_backup
        return enum_number

    @property
    def lo_source(self) -> OctaveLOSource:
        return self._client.rf_outputs[self._octave_if_input_port_number].get_lo_source()

    def set_lo_frequency(self, lo_frequency: float, set_source: bool = True):
        # TODO - validate that LO frequency can be casted to int, or it should be a float in the proto
        """
        Sets the LO frequency of the synthesizer associated to element

        :param lo_frequency:
        :param set_source:
        """
        if self.lo_source != OctaveLOSource.Internal and self.lo_source not in self._client._port_mapping:
            raise SetFrequencyException(f"Cannot set frequency to an external lo source {self.lo_source.name}")

        if set_source:
            self._client.rf_outputs[self._octave_if_input_port_number].set_lo_source(
                self.lo_source, ignore_shared_errors=True
            )

        self._client.rf_outputs[self._octave_if_input_port_number].set_lo_frequency(self.lo_source, lo_frequency)
        self.lo_frequency = lo_frequency

    def set_lo_source(self, lo_port: OctaveLOSource):
        """
        Sets the source of LO going to the upconverter associated with element.
        :param lo_port:
        """
        self._client.rf_outputs[self._octave_if_input_port_number].set_lo_source(lo_port, ignore_shared_errors=True)

    def set_rf_output_mode(self, switch_mode: RFOutputMode):
        """
        Configures the output switch of the upconverter associated to element.
        switch_mode can be either: 'always_on', 'always_off', 'normal' or 'inverted'
        When in 'normal' mode a high trigger will turn the switch on and a low trigger will turn it off
        When in 'inverted' mode a high trigger will turn the switch off and a low trigger will turn it on
        When in 'always_on' the switch will be permanently on. When in 'always_off' mode the switch will
        be permanently off.

        :param switch_mode:
        """
        self._client.rf_outputs[self._octave_if_input_port_number].set_output(switch_mode)

    def set_rf_output_gain(self, gain_in_db: float):
        """
        Sets the RF output gain for the up-converter associated with the element.
        RF_gain is in steps of 0.5dB from -20 to +24 and is referring
        to the maximum OPX output power of 4dBm (=0.5V pk-pk) So for a value of -24
        for example, an IF signal coming from the OPX at max power (4dBm) will be
        upconverted and come out of Octave at -20dBm

        :param gain_in_db:
        """
        # a. use value custom set in qm.octave.update_external
        # b. use value from config
        self._client.rf_outputs[self._octave_if_input_port_number].set_gain(gain_in_db, self.lo_frequency)

    def set_downconversion(
        self,
        lo_source: Optional[RFInputLOSource] = None,
        lo_frequency: Optional[float] = None,
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
    ):
        """
        Sets the LO source for the downconverters.
        The LO source will be the one associated with the upconversion of element

        :param lo_source:
        :param lo_frequency:
        :param if_mode_i:
        :param if_mode_q:
        """
        self._set_downconversion_lo(lo_source, lo_frequency)
        self._set_downconversion_if_mode(if_mode_i, if_mode_q)

    def _set_downconversion_lo(
        self,
        lo_source: Optional[RFInputLOSource] = None,
        lo_frequency: Optional[float] = None,
    ):
        """
        Sets the LO source for the downconverters.
        If no value is given the LO source will be the one associated with the
        upconversion of element

        :param lo_frequency:
        :param lo_source:
        """
        # TODO check if the downconversion shared between elements(from same rf_in or both rf_ins)
        # if we have two elements for the same downconversion:
        # warning when setting for port that has multiple elements
        # “element x shares the input port with elements [y],
        # any settings applied will affect all elements”
        if self._octave_rf_input_port_number is None:
            raise AttributeError("This element has no connection ot octave RF-in, cannot set downconversion")

        if lo_source is None:
            lo_source = self.lo_source

        self._downconversion_client.rf_inputs[self._octave_rf_input_port_number].set_lo_source(
            lo_source, ignore_shared_errors=True
        )
        self._downconversion_client.rf_inputs[self._octave_rf_input_port_number].set_rf_source(RFInputRFSource.RF_in)
        is_internal = lo_source in {RFInputLOSource.Internal, RFInputLOSource.Analyzer}

        if lo_frequency is not None and is_internal:
            self._downconversion_client.rf_inputs[self._octave_rf_input_port_number].set_lo_frequency(
                source_name=lo_source, frequency=lo_frequency
            )

    def _set_downconversion_if_mode(self, if_mode_i: IFMode = IFMode.direct, if_mode_q: IFMode = IFMode.direct):
        if self._octave_rf_input_port_number is None:
            raise AttributeError("This element has no connection ot octave RF-in, cannot set downconversion")
        self._downconversion_client.rf_inputs[self._octave_rf_input_port_number].set_if_mode_i(if_mode_i)
        self._downconversion_client.rf_inputs[self._octave_rf_input_port_number].set_if_mode_q(if_mode_q)

    def set_clock(self, clock_type: ClockType, frequency: Optional[ClockFrequency]):
        """
        This function will set the octave clock type - internal, external or buffered.
        It can also set the clock frequency - 10, 100 or 1000 MHz

        :param clock_type: ClockType
        :param frequency: ClockFrequency
        """
        self._client.set_clock(clock_type, frequency)
        self._client.save_default_state(only_clock=True)

    @property
    def clock(self) -> ClockInfo:
        return self._client.get_clock()



from typing import List, Tuple


class QmQuaException(Exception):
    def __init__(self, message: str, *args):
        self.message = message
        super().__init__(message, *args)


class QmmException(QmQuaException):
    pass


class OpenQmException(QmQuaException):
    def __init__(self, message: str, *args, errors: List[Tuple[str, str, str]]):
        super().__init__(message, *args)
        self.errors = errors


class FailedToExecuteJobException(QmQuaException):
    pass


class FailedToAddJobToQueueException(QmQuaException):
    pass


class CompilationException(QmQuaException):
    pass


class JobCancelledError(QmQuaException):
    pass


class ErrorJobStateError(QmQuaException):
    def __init__(self, *args, error_list: List[str]):
        super().__init__(*args)
        self._error_list = error_list if error_list else []

    def __str__(self):
        errors_string = "\n".join(error for error in self._error_list)
        return f"{super().__str__()}\n{errors_string}"


class UnknownJobStateError(QmQuaException):
    pass


class InvalidStreamMetadataError(QmQuaException):
    pass


class ConfigValidationException(QmQuaException):
    pass


class ConfigSerializationException(QmQuaException):
    pass


class UnsupportedCapabilityError(QmQuaException):
    pass


class InvalidConfigError(QmQuaException):
    pass


class QMHealthCheckError(QmQuaException):
    pass


class QMFailedToGetQuantumMachineError(QmQuaException):
    pass


class QMSimulationError(QmQuaException):
    pass


class QmFailedToCloseQuantumMachineError(QmQuaException):
    pass


class QMFailedToCloseAllQuantumMachinesError(QmFailedToCloseQuantumMachineError):
    pass


class QMRequestError(QmQuaException):
    pass


class QMConnectionError(QmQuaException):
    pass


class QMTimeoutError(QmQuaException):
    pass


class QMRequestDataError(QmQuaException):
    pass


class QmServerDetectionError(QmQuaException):
    pass


class QmValueError(QmQuaException):
    pass


class QmInvalidSchemaError(QmQuaException):
    pass


class QmInvalidResult(QmQuaException):
    pass



from typing import TypeVar, Collection, Type, Iterable, Any, Set, Union, Optional, cast, Sequence

import numpy as np
import numpy.typing

from qm.type_hinting import Value, NumpySupportedValue


def fix_object_data_type(obj: Optional[NumpySupportedValue]) -> Optional[Value]:
    if isinstance(obj, (np.floating, np.integer, np.bool_)):
        obj_item = obj.item()
        if isinstance(obj_item, np.longdouble):
            return float(obj_item)
        else:
            return obj_item
    else:
        return obj


def get_all_iterable_data_types(it: Iterable[Any]) -> Set[Type[Any]]:
    return {type(e) for e in it}


C = TypeVar("C")


def collection_has_type(collection: Collection[C], type_to_check: Type[C], include_subclasses: bool) -> bool:
    if include_subclasses:
        return any([isinstance(i, type_to_check) for i in collection])
    else:
        return any([type(i) is type_to_check for i in collection])


def collection_has_type_bool(collection: Collection[C]) -> bool:
    return collection_has_type(collection, bool, False) or collection_has_type(collection, np.bool_, True)


def collection_has_type_int(collection: Collection[C]) -> bool:
    return collection_has_type(collection, int, False) or collection_has_type(collection, np.integer, True)


def collection_has_type_float(collection: Collection[C]) -> bool:
    return collection_has_type(collection, float, False) or collection_has_type(collection, np.floating, True)


def is_iter(x: Any) -> bool:
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True


def get_iterable_elements_datatype(it: Union[numpy.typing.NDArray[Any], Sequence[Any], Any]) -> Optional[Type[Any]]:
    if isinstance(it, np.ndarray):
        return type(it[0].item())
    elif is_iter(it):
        sequence = cast(Sequence[Any], it)
        if len(get_all_iterable_data_types(sequence)) > 1:
            raise ValueError("Multiple datatypes encountered in iterable object")
        if isinstance(sequence[0], np.generic):
            return type(sequence[0].item())
        else:
            return type(sequence[0])
    else:
        return None



from octave_sdk import (
    OctaveOutput,
    ClockType,
    ClockFrequency,
    OctaveLOSource,
    IFMode,
    RFInputLOSource,
    RFInputRFSource,
    RFOutputMode,
)
from qm.octave.octave_config import QmOctaveConfig
from qm.octave.calibration_db import octave_output_mixer_name

__all__ = [
    "OctaveOutput",
    "ClockType",
    "ClockFrequency",
    "OctaveLOSource",
    "IFMode",
    "RFInputLOSource",
    "RFInputRFSource",
    "RFOutputMode",
    "octave_output_mixer_name",
    "QmOctaveConfig",
]



import warnings

from qm.program.program import Program

_Program = Program

warnings.warn(
    "'qm.program._Program' is moved as of 1.1.0 and will be removed in 1.2.0. " "use 'qm.Program' instead",
    category=DeprecationWarning,
)



import json
import logging
import warnings

from typing import Optional, List, Dict, Any

from typing_extensions import TypedDict

from qm.program import Program
from qm.simulate.interface import SimulationConfig
from qm.api.models.compiler import CompilerOptionArguments
from qm.containers.capabilities_container import create_capabilities_container
from qm.QuantumMachine import QuantumMachine
from qm._controller import Controller
from qm.api.frontend_api import FrontendApi
from qm.api.models.debug_data import DebugData
from qm.api.models.server_details import ServerDetails
from qm.api.server_detector import detect_server
from qm.api.simulation_api import SimulationApi
from qm.exceptions import QmmException
from qm.jobs.simulated_job import SimulatedJob
from qm.logging_utils import set_logging_level
from qm.octave import QmOctaveConfig
from qm.octave.calibration_db import load_from_calibration_db
from qm.octave.octave_manager import OctaveManager
from qm.persistence import SimpleFileStore, BaseStore
from qm.program import load_config
from qm.program._qua_config_schema import validate_config_capabilities
from qm.program._qua_config_to_pb import load_config_pb
from qm.type_hinting.config_types import DictQuaConfig
from qm.user_config import UserConfig
from qm.utils import deprecation_message

logger = logging.getLogger(__name__)

Version = TypedDict("Version", {"client": str, "server": Optional[str]})


class QuantumMachinesManager:
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        *,
        timeout: Optional[float] = None,
        log_level: int = logging.INFO,
        connection_headers: Optional[Dict[str, str]] = None,
        add_debug_data: bool = False,
        credentials: Optional[str] = None,
        store: Optional[BaseStore] = None,
        file_store_root: str = ".",
        octave: Optional[QmOctaveConfig] = None,
    ):
        """
        Args:
            host (string): Host where to find the QM orchestrator. If
                ``None``, local settings are used
            port: Port where to find the QM orchestrator. If None, local
                settings are used
        """
        set_logging_level(log_level)

        self._user_config = UserConfig.create_from_file()

        self._port = port
        self._host = host or self._user_config.manager_host
        if self._host is None:
            message = "Failed to connect to QuantumMachines server. No host given."
            logger.error(message)
            raise QmmException(message)

        self._credentials = credentials
        self._store = store if store else SimpleFileStore(file_store_root)

        self._octave_config = octave
        self._octave_manager = OctaveManager(octave, self)

        self._server_details = self._initialize_connection(
            timeout=timeout,
            add_debug_data=add_debug_data,
            connection_headers=connection_headers,
        )

        self._caps = self._server_details.capabilities
        self._frontend = FrontendApi(self._server_details.connection_details)
        self._simulation_api = SimulationApi(self._server_details.connection_details)

        self._octave_config = octave
        self._octave_manager = OctaveManager(self._octave_config, self, self._caps)

        raise_on_error = self._user_config.strict_healthcheck is not False
        self.perform_healthcheck(raise_on_error)

    def _initialize_connection(
        self,
        timeout: Optional[float],
        add_debug_data: bool,
        connection_headers: Optional[Dict[str, str]],
    ) -> ServerDetails:
        server_details = detect_server(
            user_token=self._user_config.user_token,
            credentials=self._credentials,
            host=self._host,
            port_from_user_config=self._user_config.manager_port,
            user_provided_port=self._port,
            add_debug_data=add_debug_data,
            timeout=timeout,
            extra_headers=connection_headers,
        )
        create_capabilities_container(server_details.qua_implementation)
        return server_details

    @property
    def store(self) -> BaseStore:
        return self._store

    @property
    def octave_manager(self) -> OctaveManager:
        warnings.warn("Do not use OctaveManager, it will be removed in the next version", DeprecationWarning)
        return self._octave_manager

    def perform_healthcheck(self, strict: bool = True) -> None:
        """Perform a health check against the QM server.

        Args:
            strict: Will raise an exception if health check failed
        """
        self._frontend.healthcheck(strict)

    def version(self) -> Version:
        """
        Returns:
            The SDK and QOP versions
        """
        server_version = self._server_details.qop_version
        from qm.version import __version__

        return {"client": __version__, "server": server_version}

    def reset_data_processing(self) -> None:
        """Stops current data processing for ALL running jobs"""
        self._frontend.reset_data_processing()

    def close(self) -> None:
        """Closes the Quantum machine manager"""
        warnings.warn(
            deprecation_message("QuantumMachineManager.close()", "1.1.0", "1.2.0", "close will be removed."),
            category=DeprecationWarning,
        )
        pass

    def open_qm(
        self,
        config: DictQuaConfig,
        close_other_machines: bool = True,
        validate_with_protobuf: bool = False,
        use_calibration_data: bool = True,
        **kwargs: Any,
    ) -> QuantumMachine:
        """Opens a new quantum machine. A quantum machine can use multiple OPXes, and a
        single OPX can also be used by multiple quantum machines as long as they do not
        share the same physical resources (input/output ports) as defined in the config.

        Args:
            config: The config that will be used by the quantum machine
            close_other_machines: When set to true (default) any open
                quantum machines will be closed. This simplifies the
                workflow, but does not enable opening more than one
                quantum machine.
            validate_with_protobuf (bool): Validates config with
                protobuf instead of marshmallow. It is usually faster
                when working with large configs. Defaults to False.

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        if kwargs:
            logger.warning(f"unused kwargs: {list(kwargs)}, please remove them.")

        if use_calibration_data and self._octave_config is not None and self._octave_config.calibration_db is not None:
            load_from_calibration_db(config, self._octave_config.calibration_db)

        if validate_with_protobuf:
            loaded_config = load_config_pb(config)
        else:
            loaded_config = load_config(config)
        validate_config_capabilities(loaded_config, self._caps)

        machine_id = self._frontend.open_qm(loaded_config, close_other_machines)

        return QuantumMachine(
            machine_id=machine_id,
            pb_config=loaded_config,
            frontend_api=self._frontend,
            capabilities=self._caps,
            store=self._store,
            octave_config=self._octave_config,
            octave_manager=self._octave_manager,
        )

    def open_qm_from_file(self, filename: str, close_other_machines: bool = True) -> QuantumMachine:
        """Opens a new quantum machine with config taken from a file on the local file system

        Args:
            filename: The path to the file that contains the config
            close_other_machines: Flag whether to close all other
                running machines

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        with open(filename) as json_file:
            json1_str = json_file.read()

            def remove_nulls(d: Dict[Any, Any]) -> Dict[Any, Any]:
                return {k: v for k, v in d.items() if v is not None}

            config = json.loads(json1_str, object_hook=remove_nulls)
        return self.open_qm(config, close_other_machines)

    def simulate(
        self,
        config: DictQuaConfig,
        program: Program,
        simulate: SimulationConfig,
        compiler_options: Optional[CompilerOptionArguments] = None,
    ) -> SimulatedJob:
        """Simulate the outputs of a deterministic QUA program.

        The following example shows a simple execution of the simulator, where the
        associated config object is omitted for brevity.

        Example:
            ```python
            from qm.QuantumMachinesManager import QuantumMachinesManager
            from qm.qua import *
            from qm import SimulationConfig

            qmm = QuantumMachinesManager()

            with program() as prog:
                play('pulse1', 'qe1')

            job = qmm.simulate(config, prog, SimulationConfig(duration=100))
            ```
        Args:
            config: A QM config
            program: A QUA ``program()`` object to execute
            simulate: A ``SimulationConfig`` configuration object
            kwargs: additional parameters to pass to execute

        Returns:
            a ``QmJob`` object (see QM Job API).
        """
        if compiler_options is None:
            compiler_options = CompilerOptionArguments()

        job_id, simulated_response_part = self._simulation_api.simulate(config, program, simulate, compiler_options)
        return SimulatedJob(
            job_id=job_id,
            frontend_api=self._frontend,
            capabilities=self._server_details.capabilities,
            store=self._store,
            simulated_response=simulated_response_part,
        )

    def list_open_quantum_machines(self) -> List[str]:
        """Return a list of open quantum machines. (Returns only the ids, use ``get_qm(...)`` to get the machine object)

        Returns:
            The ids list
        """
        return self._frontend.list_open_quantum_machines()

    def get_qm(self, machine_id: str) -> QuantumMachine:
        """Gets an open quantum machine object with the given machine id

        Args:
            machine_id: The id of the open quantum machine to get

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        qm_data = self._frontend.get_quantum_machine(machine_id)
        return QuantumMachine(
            machine_id=qm_data.machine_id,
            pb_config=qm_data.config,
            frontend_api=self._frontend,
            capabilities=self._caps,
            store=self.store,
            octave_manager=self.octave_manager,
        )

    def close_all_quantum_machines(self) -> None:
        """Closes ALL open quantum machines"""
        self._frontend.close_all_quantum_machines()

    def get_controllers(self) -> List[Controller]:
        """Returns a list of all the controllers that are available"""
        return [Controller.build_from_message(message) for message in self._frontend.get_controllers()]

    def clear_all_job_results(self) -> None:
        """Deletes all data from all previous jobs"""
        self._frontend.clear_all_job_results()

    def _send_debug_command(self, controller_name: str, command: str) -> str:
        return self._frontend.send_debug_command(controller_name, command)

    @property
    def _debug_data(self) -> Optional[DebugData]:
        return self._server_details.connection_details.debug_data



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/grpc/quantum_simulator/v1/physical_system.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import (
    Dict,
    List,
)

import betterproto


@dataclass(eq=False, repr=False)
class InputPort(betterproto.Message):
    """Input port of the quantum system (output port of OPX)"""

    controller: str = betterproto.string_field(1)
    """name of the controller"""

    id: int = betterproto.uint32_field(2)
    """id"""


@dataclass(eq=False, repr=False)
class OutputPort(betterproto.Message):
    """Output port of the quantum system (input port of OPX)"""

    controller: str = betterproto.string_field(1)
    """name of the controller"""

    id: int = betterproto.uint32_field(2)
    """id"""


@dataclass(eq=False, repr=False)
class TransmonProperties(betterproto.Message):
    """A class that holds the physical properties of a transmon."""

    frequency: float = betterproto.double_field(1)
    """frequency of the qubit in Hz"""

    anharmonicity: float = betterproto.double_field(2)
    """anharmonicity of the qubit in Hz"""

    t1: float = betterproto.double_field(3)
    """T1 time in ns"""

    t2: float = betterproto.double_field(4)
    """T2 time in ns"""

    n_levels: int = betterproto.uint32_field(5)
    n_order: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class IqMixerIn(betterproto.Message):
    """IQMixer for up conversion"""

    i: "InputPort" = betterproto.message_field(1)
    """output port of the OPX"""

    q: "InputPort" = betterproto.message_field(2)
    """output port of the OPX"""

    lo_frequency: float = betterproto.double_field(3)
    """frequency of the local oscillator in Hz"""


@dataclass(eq=False, repr=False)
class IqMixerOut(betterproto.Message):
    """IQMixer for down conversion"""

    i: "OutputPort" = betterproto.message_field(1)
    """input port of the OPX"""

    q: "OutputPort" = betterproto.message_field(2)
    """input port of the OPX"""

    lo_frequency: float = betterproto.double_field(3)
    """frequency of the local oscillator in Hz"""


@dataclass(eq=False, repr=False)
class AnalogFrontend(betterproto.Message):
    """
    A class to describe analog frontend (currently just a constant
    multiplicative factor)
    """

    amp_factor: float = betterproto.double_field(1)
    """
    a constant factor to account for losses or amplification of the signal
    """


@dataclass(eq=False, repr=False)
class FluxLine(betterproto.Message):
    """
    A class to describe flux line and how it is connected to the OPX. The
    frequency of the qubit follows is tuned as: amp*(flux voltage - offset)^2
    """

    port: "InputPort" = betterproto.message_field(1)
    """output port of the OPX driving the flux line"""

    amp_factor: float = betterproto.double_field(2)
    """amplitude in units of GHz/V^2"""

    offset: float = betterproto.double_field(3)
    """offset voltage at which the bias is zero"""


@dataclass(eq=False, repr=False)
class FixedFrequencyTransmon(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class TunableFrequencyTransmon(betterproto.Message):
    """A class to describe tunable frequency transmon"""

    flux_line: "FluxLine" = betterproto.message_field(1)
    """flux line controlling the transmon frequency"""


@dataclass(eq=False, repr=False)
class Transmon(betterproto.Message):
    """
    Transmon contains the physical properties of the transmonand the
    information about how it is connected to the IQ Mixer and OPX
    """

    properties: "TransmonProperties" = betterproto.message_field(1)
    """physical properties of the transmon"""

    i_q_mixer: "IqMixerIn" = betterproto.message_field(2)
    """IQ mixer used to drive the qubit"""

    analog_frontend: "AnalogFrontend" = betterproto.message_field(3)
    """analog frontend"""

    fixed_frequency_transmon: "FixedFrequencyTransmon" = betterproto.message_field(
        4, group="type"
    )
    """fixed frequency qubit"""

    tunable_frequency_transmon: "TunableFrequencyTransmon" = betterproto.message_field(
        5, group="type"
    )
    """tunable frequency qubit"""


@dataclass(eq=False, repr=False)
class DriveResonator(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ReadoutResonator(betterproto.Message):
    """
    A class with the information about how the back reflected from the
    resonator is fed to the OPX.
    """

    i_q_mixer: "IqMixerOut" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResonatorProperties(betterproto.Message):
    frequency: float = betterproto.double_field(1)
    """resonator frequency in Hz"""

    lossrate: float = betterproto.double_field(2)
    """resonator loss rate in Hz"""

    n_levels: int = betterproto.uint32_field(3)
    """number of n_levels"""


@dataclass(eq=False, repr=False)
class Resonator(betterproto.Message):
    """
    A class to describe the physical properties of readout resonator and how it
    is connected to the OPX
    """

    properties: "ResonatorProperties" = betterproto.message_field(1)
    i_q_mixer: "IqMixerIn" = betterproto.message_field(2)
    """IQMixer to upconvert the drive"""

    analog_frontend: "AnalogFrontend" = betterproto.message_field(3)
    """analog frontend"""

    drive_resonator: "DriveResonator" = betterproto.message_field(4, group="type")
    """drive resonator"""

    readout_resonator: "ReadoutResonator" = betterproto.message_field(5, group="type")
    """readout resonator"""


@dataclass(eq=False, repr=False)
class CapacitiveCoupler(betterproto.Message):
    """A class for describing capacitive coupling between two qubits"""

    g: "InputPort" = betterproto.message_field(1)
    """output port of the OPX driving the coupler"""

    amp_factor: float = betterproto.double_field(2)
    """amplitude factor in units of GHz/V"""


@dataclass(eq=False, repr=False)
class TransmonTransmonCoupling(betterproto.Message):
    """Transmon-transmon coupling"""

    capacitive_coupler: str = betterproto.string_field(1)
    """name of the capacitive coupler"""

    qb1: str = betterproto.string_field(2)
    """name of the qubit 1"""

    qb2: str = betterproto.string_field(3)
    """name of the qubit 2"""


@dataclass(eq=False, repr=False)
class TransmonResonatorCoupling(betterproto.Message):
    """
    Transmon-resonator coupling describes the connectivity between a qubit and
    a resonator
    """

    transmon: str = betterproto.string_field(1)
    """name of the qubit"""

    resonator: str = betterproto.string_field(2)
    """name of the resonator"""

    g: float = betterproto.double_field(3)
    """strength of the coupling in GHz"""

    is_dispersive: bool = betterproto.bool_field(4)
    """coupling type"""


@dataclass(eq=False, repr=False)
class ScChip(betterproto.Message):
    """
    A class to describe a superconducting chip layout consisting of fixed/flux
    tunable transmons, capacitive couplers and readout resonators.
    """

    transmons: Dict[str, "Transmon"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """a dictionary of Transmons"""

    resonators: Dict[str, "Resonator"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """a dictionary of resonators"""

    capacitive_couplers: Dict[str, "CapacitiveCoupler"] = betterproto.map_field(
        3, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """a dictionary of capacitive couplers"""

    transmon_transmon_coupling: List[
        "TransmonTransmonCoupling"
    ] = betterproto.message_field(4)
    """a list containing all the qubit-qubit couplings in the chip"""

    transmon_resonator_coupling: List[
        "TransmonResonatorCoupling"
    ] = betterproto.message_field(5)
    """
    a list containing the qubit-resonator couplings and their respective
    strengths
    """


@dataclass(eq=False, repr=False)
class Offset(betterproto.Message):
    port: "InputPort" = betterproto.message_field(1)
    value: float = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class Waveform(betterproto.Message):
    constant_waveform: "ConstantWaveform" = betterproto.message_field(1, group="type")
    arbitrary_waveform: "ArbitraryWaveform" = betterproto.message_field(2, group="type")


@dataclass(eq=False, repr=False)
class ConstantWaveform(betterproto.Message):
    sample: float = betterproto.double_field(1)


@dataclass(eq=False, repr=False)
class ArbitraryWaveform(betterproto.Message):
    samples: List[float] = betterproto.double_field(1)


@dataclass(eq=False, repr=False)
class Error(betterproto.Message):
    type: str = betterproto.string_field(1)
    stack: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ColdAtomDevice(betterproto.Message):
    v: float = betterproto.double_field(1)
    qubit: str = betterproto.string_field(2)
    frequency_01: float = betterproto.double_field(3)
    frequency_1_r: float = betterproto.double_field(4)
    positions: List["Coordinate"] = betterproto.message_field(5)
    t1: float = betterproto.double_field(6)
    t2: float = betterproto.double_field(7)


@dataclass(eq=False, repr=False)
class Coordinate(betterproto.Message):
    x: float = betterproto.double_field(1)
    y: float = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class ColdAtomSetup(betterproto.Message):
    device: "ColdAtomDevice" = betterproto.message_field(1)
    global_raman_beam: "GlobalRamanBeam" = betterproto.message_field(2)
    global_rydberg_beam: "GlobalRydbergBeam" = betterproto.message_field(3)
    local_raman_detuning: "AcStarkShifter" = betterproto.message_field(4)
    local_rydberg_detuning: "AcStarkShifter" = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class BeamSteerer(betterproto.Message):
    aod_x: str = betterproto.string_field(1)
    aod_y: str = betterproto.string_field(2)
    reference_position: "Coordinate" = betterproto.message_field(3)
    eta_x: float = betterproto.double_field(4)
    eta_y: float = betterproto.double_field(5)


@dataclass(eq=False, repr=False)
class AcStarkShifter(betterproto.Message):
    aom: str = betterproto.string_field(1)
    delta: float = betterproto.double_field(2)
    eta_x: List[float] = betterproto.double_field(3)
    beam_steerer: "BeamSteerer" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class GlobalRamanBeam(betterproto.Message):
    intensity_modulator: str = betterproto.string_field(1)
    frequency: float = betterproto.double_field(2)
    scale_factor: float = betterproto.double_field(3)
    lo_frequency: float = betterproto.double_field(4)


@dataclass(eq=False, repr=False)
class GlobalRydbergBeam(betterproto.Message):
    aom1: str = betterproto.string_field(1)
    aom2: str = betterproto.string_field(2)
    frequency1: float = betterproto.double_field(3)
    frequency2: float = betterproto.double_field(4)
    scale_factor1: float = betterproto.double_field(5)
    scale_factor2: float = betterproto.double_field(6)


@dataclass(eq=False, repr=False)
class PhysicalConfig(betterproto.Message):
    chip: "ScChip" = betterproto.message_field(1, group="type")
    cold_atom_setup: "ColdAtomSetup" = betterproto.message_field(2, group="type")



from qm.utils.types_utils import (
    fix_object_data_type,
    get_all_iterable_data_types,
    collection_has_type,
    collection_has_type_bool,
    collection_has_type_int,
    collection_has_type_float,
    is_iter,
    get_iterable_elements_datatype,
)

from qm.utils.deprecation_utils import deprecate_to_property, deprecation_message

from qm.utils.protobuf_utils import LOG_LEVEL_MAP, list_fields

from qm.utils.general_utils import run_until_with_timeout

__all__ = [
    "LOG_LEVEL_MAP",
    "fix_object_data_type",
    "get_all_iterable_data_types",
    "collection_has_type",
    "collection_has_type_bool",
    "collection_has_type_int",
    "collection_has_type_float",
    "is_iter",
    "get_iterable_elements_datatype",
    "deprecate_to_property",
    "deprecation_message",
    "list_fields",
    "run_until_with_timeout",
]



import os
import logging
from dataclasses import dataclass, asdict
from pathlib import Path
from typing import List, Dict, Union, Any, Optional

from tinydb import TinyDB, Query

from qm.type_hinting.config_types import DictQuaConfig

logger = logging.getLogger(__name__)


@dataclass
class CalibrationResult:
    correction: List[float]
    i_offset: float
    q_offset: float
    lo_frequency: float
    if_frequency: float
    temperature: float
    mixer_id: str
    optimizer_parameters: Dict[str, Any]


class CalibrationDB:
    def __init__(self, path: Union[str, Path]) -> None:
        super().__init__()
        file_name = "calibration_db.json"
        self._file_path = os.path.join(path, file_name)
        self._db = TinyDB(self._file_path)

    @property
    def file_path(self):
        return self._file_path

    def update_calibration_data(self, data: Union[CalibrationResult, List[CalibrationResult]]):
        if isinstance(data, CalibrationResult):
            data = [data]

        query = Query()
        for result in data:
            self._db.upsert(
                asdict(result),
                (query.mixer_id == result.mixer_id)
                & (query.lo_frequency == result.lo_frequency)
                & (query.if_frequency == result.if_frequency),
            )

    def get(self, mixer, lo_freq, if_freq) -> Optional[CalibrationResult]:
        query = Query()
        doc = self._db.get(
            (query.mixer_id == mixer) & (query.lo_frequency == lo_freq) & (query.if_frequency == if_freq),
        )
        if doc is not None:
            return self._doc_to_result(doc)
        else:
            raise AttributeError(f"calibration for {mixer}, lo {lo_freq}, if {if_freq} was not found")

    def get_or_none(self, mixer, lo_freq, if_freq) -> Optional[CalibrationResult]:
        query = Query()
        doc = self._db.get(
            (query.mixer_id == mixer) & (query.lo_frequency == lo_freq) & (query.if_frequency == if_freq)
        )
        if doc is not None:
            return self._doc_to_result(doc)
        else:
            return None

    def get_for_lo_frequency(self, mixer, lo_freq) -> List[CalibrationResult]:
        query = Query()
        table = self._db.search((query.mixer_id == mixer) & (query.lo_frequency == lo_freq))
        return [self._doc_to_result(doc) for doc in table]

    def get_all(self, mixer) -> List[CalibrationResult]:
        l = self.get_all_or_default(mixer)
        if len(l) > 0:
            return l
        else:
            raise AttributeError(f"calibration for {mixer} was not found")

    def get_all_or_default(self, mixer) -> Optional[List[CalibrationResult]]:
        query = Query()
        table = self._db.search(query.mixer_id == mixer)
        l = [self._doc_to_result(doc) for doc in table]
        return l

    def __getitem__(self, item):
        return self.get(*item)

    @staticmethod
    def _doc_to_result(doc):
        return CalibrationResult(
            correction=doc["correction"],
            i_offset=doc["i_offset"],
            q_offset=doc["q_offset"],
            if_frequency=doc["if_frequency"],
            lo_frequency=doc["lo_frequency"],
            temperature=doc["temperature"],
            optimizer_parameters=doc["optimizer_parameters"],
            mixer_id=doc["mixer_id"],
        )

    def __del__(self):
        self._db.close()


def octave_output_mixer_name(octave: str, port: int) -> str:
    return f"octave_{octave}_{port}"


def load_from_calibration_db(config: DictQuaConfig, calibration_db: CalibrationDB) -> None:
    if calibration_db is None:
        return
    for mixer in config["mixers"].keys():
        cal_list = calibration_db.get_all_or_default(mixer)
        cal_dict = {(v.lo_frequency, v.if_frequency): v.correction for v in cal_list}
        if len(cal_list) > 0:
            old_values = config["mixers"][mixer]
            new_values = [
                {
                    "intermediate_frequency": int(v.if_frequency),
                    "lo_frequency": int(v.lo_frequency),
                    "correction": v.correction,
                }
                for v in cal_list
            ]

            for v in old_values:
                if (v["lo_frequency"], v["intermediate_frequency"]) in cal_dict:
                    pass
                else:
                    new_values.append(v)
                    logger.debug(
                        f"Could not find calibration value for lo frequency"
                        f" {v['lo_frequency']} and intermediate_frequency"
                        f" {v['intermediate_frequency']}"
                    )

            config["mixers"][mixer] = new_values

            # changing offsets here because we know the mixer is octave
            for element in config["elements"].values():
                if (
                    "mixInputs" in element
                    and element["mixInputs"]["mixer"] == mixer
                    and "intermediate_frequency" in element
                ):
                    if_freq = element["intermediate_frequency"]
                    lo_freq = element["mixInputs"]["lo_frequency"]
                    i_port = element["mixInputs"]["I"]
                    q_port = element["mixInputs"]["Q"]
                    result = calibration_db.get_or_none(mixer, lo_freq, if_freq)
                    if result is not None:
                        config["controllers"][i_port[0]]["analog_outputs"][i_port[1]]["offset"] = result.i_offset
                        config["controllers"][q_port[0]]["analog_outputs"][q_port[1]]["offset"] = result.q_offset



import logging
from typing import Tuple, AsyncIterator

from qm.type_hinting.config_types import DictQuaConfig
from qm.utils.async_utils import run_async
from qm.simulate import interface
from qm.simulate.interface import SimulationConfig
from qm.program import Program
from qm.api.base_api import BaseApi
from qm.api.models.compiler import (
    CompilerOptionArguments,
    _get_request_compiler_options,
)
from qm.api.models.server_details import ConnectionDetails
from qm.exceptions import QMSimulationError, FailedToExecuteJobException
from qm.grpc.frontend import (
    FrontendStub,
    SimulatedResponsePart,
    SimulationRequest,
    ExecutionRequestSimulate,
    InterOpxAddress,
    InterOpxConnection,
    InterOpxConnectionAddressToAddress,
    InterOpxChannel,
    InterOpxConnectionChannelToChannel,
    GetSimulatedQuantumStateRequest,
    DensityMatrix,
)
from qm.grpc.results_analyser import (
    PullSimulatorSamplesRequest,
    SimulatorSamplesResponse,
)
from qm.program import load_config
from qm.utils.protobuf_utils import LOG_LEVEL_MAP

logger = logging.getLogger(__name__)


class SimulationApi(BaseApi):
    def __init__(self, connection_details: ConnectionDetails):
        super().__init__(connection_details)
        self._stub = FrontendStub(self._channel)
        self._timeout = None

    def simulate(
        self,
        config: DictQuaConfig,
        program: Program,
        simulate: SimulationConfig,
        compiler_options: CompilerOptionArguments,
    ) -> Tuple[str, SimulatedResponsePart]:

        if type(program) is not Program:
            raise Exception("program argument must be of type qm.program.Program")

        request = SimulationRequest()
        msg_config = load_config(config)
        request.config = msg_config

        if type(simulate) is SimulationConfig:
            request.simulate = ExecutionRequestSimulate(
                duration=simulate.duration,
                include_analog_waveforms=simulate.include_analog_waveforms,
                include_digital_waveforms=simulate.include_digital_waveforms,
                extra_processing_timeout_ms=simulate.extraProcessingTimeoutInMs,
            )
            request = simulate.update_simulate_request(request)

            for connection in simulate.controller_connections:
                if not isinstance(connection.source, type(connection.target)):
                    raise Exception(
                        f"Unsupported InterOpx connection. Source is "
                        f"{type(connection.source).__name__} but target is "
                        f"{type(connection.target).__name__}"
                    )

                if isinstance(connection.source, interface.InterOpxAddress):
                    con = InterOpxConnection(
                        address_to_address=InterOpxConnectionAddressToAddress(
                            source=InterOpxAddress(
                                controller=connection.source.controller,
                                left=connection.source.is_left_connection,
                            ),
                            target=InterOpxAddress(
                                controller=connection.target.controller,
                                left=connection.target.is_left_connection,
                            ),
                        )
                    )
                elif isinstance(connection.source, interface.InterOpxChannel):
                    con = InterOpxConnection(
                        channel_to_channel=InterOpxConnectionChannelToChannel(
                            source=InterOpxChannel(
                                controller=connection.source.controller,
                                channel_number=connection.source.channel_number,
                            ),
                            target=InterOpxChannel(
                                controller=connection.target.controller,
                                channel_number=connection.target.channel_number,
                            ),
                        )
                    )
                else:
                    raise Exception(
                        f"Unsupported InterOpx connection. Source is "
                        f"{type(connection.source).__name__}. Supported types are "
                        f"InterOpxAddress "
                        f"or InterOpxChannel"
                    )

                request.controller_connections.append(con)

        request.high_level_program = program.build(msg_config)
        request.high_level_program.compiler_options = _get_request_compiler_options(compiler_options)

        logger.info("Simulating program")

        response = run_async(self._stub.simulate(request, timeout=self._timeout))

        messages = [(LOG_LEVEL_MAP[msg.level], msg.message) for msg in response.messages]

        config_messages = [(LOG_LEVEL_MAP[msg.level], msg.message) for msg in response.config_validation_errors]

        job_id = response.job_id

        for lvl, msg in messages:
            logger.log(lvl, msg)

        for lvl, msg in config_messages:
            logger.log(lvl, msg)

        if not response.success:
            logger.error("Job " + job_id + " failed. Failed to execute program.")
            for error in response.simulated.errors:
                logger.error(f"Simulation error: {error}")
            raise FailedToExecuteJobException(job_id)

        return job_id, response.simulated

    def get_simulated_quantum_state(self, job_id: str) -> DensityMatrix:
        request = GetSimulatedQuantumStateRequest(job_id=job_id)
        response = run_async(self._stub.get_simulated_quantum_state(request, timeout=self._timeout))

        if response.ok:
            return response.state

        raise QMSimulationError("Error while pulling quantum state")

    def pull_simulator_samples(
        self, job_id: str, include_analog: bool, include_digital: bool
    ) -> AsyncIterator[SimulatorSamplesResponse]:
        request = PullSimulatorSamplesRequest(
            job_id=job_id,
            include_analog=include_analog,
            include_digital=include_digital,
            include_all_connections=True,
        )

        return self._stub.pull_simulator_samples(request)



from typing import Tuple, List, Union, TypeVar, TYPE_CHECKING, Type
from qm.grpc import qua
import numpy as np
import numpy.typing as npt

if TYPE_CHECKING:
    from _dsl import _Expression, _Variable, fixed, _ResultSource  # noqa

MessageExpressionType = qua.QuaProgramAnyScalarExpression
MessageArrayVarType = qua.QuaProgramArrayVarRefExpression
MessageVarType = qua.QuaProgramVarRefExpression
MessageVariableType = Union[MessageArrayVarType, MessageVarType]
MessageVariableOrExpression = Union[MessageExpressionType, MessageVariableType]

ArrayT = TypeVar("ArrayT")
PyArrayType = Union[List[ArrayT], npt.NDArray[ArrayT], Tuple[ArrayT, ...]]

NpTypes = Union[np.floating, np.integer, np.bool_, np.longdouble]

PyIntType = Union[int, np.integer]
PyFloatType = Union[float, np.floating, np.longdouble]
PyNumberType = Union[PyFloatType, PyIntType]
PyNumberArrayType = PyArrayType[PyNumberType]
PyBoolType = Union[bool, np.bool_]

AllPyTypes = Union[PyBoolType, PyNumberType]

QuaExpressionType = Union["_Expression", MessageExpressionType]
QuaVariableType = Union["_Variable", MessageVariableType]

QuaGenericT = TypeVar("QuaGenericT")
QuaGenericType = Union[QuaGenericT, QuaVariableType]

QuaArrayT = TypeVar("QuaArrayT")
QuaArrayType = Union[PyArrayType[QuaArrayT], QuaVariableType]

QuaIntType = QuaGenericType[PyIntType]
QuaBoolType = QuaGenericType[PyBoolType]
QuaFloatType = QuaGenericType[PyFloatType]
QuaNumberType = QuaGenericType[PyNumberType]
QuaNumberArrayType = QuaArrayType[PyNumberType]

AllQuaTypes = Union[QuaNumberType, QuaBoolType, QuaExpressionType]

VariableDeclarationType = Union[Type[int], Type[bool], Type[float], Type["fixed"]]


StreamType = Union[str, "_ResultSource"]
ChirpType = Union[
    Tuple[QuaNumberArrayType, str],
    Tuple[QuaNumberType, str],
    Tuple[QuaNumberArrayType, QuaNumberArrayType, str],
]
AmpValuesType = Tuple[
    MessageExpressionType,
    MessageExpressionType,
    MessageExpressionType,
    MessageExpressionType,
]

MeasurePulseType = Union[str, Tuple[str, AmpValuesType]]
PlayPulseType = Union[MeasurePulseType, qua.QuaProgramRampPulse]

MeasureProcessType = Union[
    "AnalogMeasureProcess.AnalogMeasureProcess",
    "DigitalMeasureProcess.DigitalMeasureProcess",
]
TimeDivisionType = Union[
    "AnalogMeasureProcess.SlicedAnalogTimeDivision",
    "AnalogMeasureProcess.AccumulatedAnalogTimeDivision",
    "AnalogMeasureProcess.MovingWindowAnalogTimeDivision",
]
AnalogProcessTargetType = Union[
    "AnalogMeasureProcess.ScalarProcessTarget",
    "AnalogMeasureProcess.VectorProcessTarget",
]

ForEachValuesType = Union[PyNumberArrayType, Tuple[PyNumberArrayType, ...]]

T = TypeVar("T")
OneOrMore = Union[T, PyArrayType[T]]

E = TypeVar("E")
TypeOrExpression = Union[E, QuaExpressionType]



from qm.jobs.job_queue import QmQueue  # noqa
import warnings

warnings.warn(
    "'qm.QmQueue.QmQueue' is moved as of 1.1.0 and will be removed in 1.2.0. " "use 'qm.QmQueue' instead",
    category=DeprecationWarning,
)



from qm.serialization.qua_serializing_visitor import QuaNodeVisitor  # noqa
import warnings

warnings.warn(
    "'qm.QuaNodeVisitor.QuaNodeVisitor' is moved as of 1.1.0 and will be removed in 1.2.0. "
    "use 'qm.serialization.qua_serializing_visitor.QuaNodeVisitor' instead",
    category=DeprecationWarning,
)



import logging
import zipfile
from io import BufferedWriter, BytesIO
from typing import Optional, Union, Dict, Tuple, Generator, List

from qm.StreamMetadata import (
    StreamMetadata,
    StreamMetadataError,
)
from qm.utils.async_utils import run_async
from qm.utils.general_utils import run_until_with_timeout
from qm.api.job_result_api import JobResultServiceApi
from qm.api.models.capabilities import ServerCapabilities
from qm.persistence import BaseStore
from qm.results.base_streaming_result_fetcher import (
    BaseStreamingResultFetcher,
    _parse_dtype,
    JobResultItemSchema,
    JobResultSchema,
)
from qm.results.multiple_streaming_result_fetcher import MultipleStreamingResultFetcher
from qm.results.single_streaming_result_fetcher import SingleStreamingResultFetcher

logger = logging.getLogger(__name__)


class StreamingResultFetcher:
    """Access to the results of a QmJob

    This object is created by calling [QmJob.result_handles][qm.jobs.running_qm_job.RunningQmJob.result_handles]

    Assuming you have an instance of StreamingResultFetcher:
    ```python
        job_results: StreamingResultFetcher
    ```
    This object is iterable:

    ```python
        for name, handle in job_results:
            print(name)
    ```

    Can detect if a name exists:

    ```python
    if "somename" in job_results:
        print("somename exists!")
        handle = job_results.get("somename")
    ```
    """

    def __init__(
        self,
        job_id: str,
        service: JobResultServiceApi,
        store: BaseStore,
        capabilities: ServerCapabilities,
    ) -> None:
        self._job_id = job_id
        self._service = service
        self._store = store
        self._schema = None
        self._capabilities = capabilities

        self._all_results: Dict[str, BaseStreamingResultFetcher] = {}
        self._add_job_results()

    def _add_job_results(self):
        self._schema = StreamingResultFetcher._load_schema(self._job_id, self._service)
        stream_metadata_errors, stream_metadata_dict = self._get_stream_metadata()
        for (name, item_schema) in self._schema.items.items():
            stream_metadata = stream_metadata_dict.get(name)
            if item_schema.is_single:
                result = SingleStreamingResultFetcher(
                    job_id=self._job_id,
                    schema=item_schema,
                    service=self._service,
                    store=self._store,
                    stream_metadata_errors=stream_metadata_errors,
                    stream_metadata=stream_metadata,
                    capabilities=self._capabilities,
                )
            else:
                result = MultipleStreamingResultFetcher(
                    job_results=self,
                    job_id=self._job_id,
                    schema=item_schema,
                    service=self._service,
                    store=self._store,
                    stream_metadata_errors=stream_metadata_errors,
                    stream_metadata=stream_metadata,
                    capabilities=self._capabilities,
                )
            self._all_results[name] = result

    def __getitem__(self, item: str):
        return self.get(item)

    def __getattr__(self, item: str):
        if item == "shape" or item == "__len__":
            return  # this is here because of a bug in pycharm debugger: ver: 2022.3.2 build #PY-223.8617.48 (24/1/23)
        return self.get(item)

    def _get_stream_metadata(
        self,
    ) -> Optional[Tuple[List[StreamMetadataError], Dict[str, StreamMetadata]]]:
        return self._service.get_program_metadata(job_id=self._job_id)

    def __iter__(self) -> Generator[Tuple[str, BaseStreamingResultFetcher], None, None]:
        for item in self._schema.items.values():
            yield item.name, self.get(item.name)

    def is_processing(self) -> bool:
        """Check if the job is still processing results

        Returns:
            True if results are still being processed, False otherwise
        """
        key = list(self._all_results.keys())[0]
        return self._all_results[key].is_processing()

    def save_to_store(
        self,
        writer: Optional[Union[BufferedWriter, BytesIO, str]] = None,
        flat_struct: bool = False,
    ):
        """Save all results to store (file system by default) in a single NPZ file

        Args:
            writer: An optional writer to be used instead of the pre-
                populated store passed to [qm.QuantumMachinesManager.QuantumMachinesManager][]
            flat_struct: results will have a flat structure - dimensions
                will be part of the shape and not of the type

        """
        own_writer = False
        if writer is None:
            own_writer = True
            writer = self._store.all_job_results(self._job_id).for_writing()
        zipf = None
        try:
            zipf = zipfile.ZipFile(writer, allowZip64=True, mode="w", compression=zipfile.ZIP_DEFLATED)
            for (name, result) in self:
                with zipf.open(f"{name}.npy", "w") as entry:
                    result.save_to_store(entry, flat_struct)
                pass
        finally:
            zipf.close()
            if own_writer:
                writer.close()

    @staticmethod
    def _load_schema(job_id: str, service: JobResultServiceApi) -> JobResultSchema:
        response = service.get_job_result_schema(job_id)
        return JobResultSchema(
            {
                item.name: JobResultItemSchema(
                    item.name,
                    _parse_dtype(item.simple_d_type),
                    tuple(item.shape),
                    item.is_single,
                    item.expected_count,
                )
                for item in response.items
            }
        )

    def get(self, name: str) -> Optional[Union[MultipleStreamingResultFetcher, SingleStreamingResultFetcher]]:
        """Get a handle to a named result from [stream_processing][qm.qua._dsl.stream_processing]

        Args:
            name: The named result using in [stream_processing][qm.qua._dsl.stream_processing]


        Returns:
            A handle object to the results `MultipleNamedJobResult` or `SingleNamedJobResult` or None if the named results in unknown
        """
        return self._all_results[name] if name in self._all_results else None

    def __contains__(self, name: str):
        return name in self._all_results

    def wait_for_all_values(self, timeout: Optional[float] = None) -> bool:
        """Wait until we know all values were processed for all named results

        Args:
            timeout: Timeout for waiting in seconds

        Returns:
            True if all finished successfully, False if any result was closed before done
        """

        def on_iteration() -> bool:
            all_job_states = [result.get_job_state() for result in self._all_results.values()]
            all_done = all(state.done for state in all_job_states)
            any_closed = any(state.closed for state in all_job_states)
            return all_done or any_closed

        def on_complete() -> bool:
            return all(result.get_job_state().done for result in self._all_results.values())

        return run_until_with_timeout(
            on_iteration_callback=on_iteration,
            on_complete_callback=on_complete,
            timeout=timeout if timeout else float("infinity"),
            timeout_message="Job was not done in time",
        )

    def get_debug_data(self, writer: Optional[Union[BufferedWriter, BytesIO, str]] = None):
        """
        Returns:
            debugging data to report to QM
        """
        if writer is None:
            writer = f"./{self._job_id}-DebugData.zip"

        owning_writer = False
        if type(writer) is str:
            writer = open(writer, "wb+")
            owning_writer = True

        try:
            run_async(self._fetch_all_job_debug_data(writer))

        finally:
            if owning_writer:
                writer.close()

    async def _fetch_all_job_debug_data(self, writer):
        async for result in self._service.get_job_debug_data(self._job_id):
            writer.write(result)



import dataclasses
import logging
from typing import Union, Dict, Iterable

import betterproto

from qm.grpc.general_messages import MessageLevel
from qm.type_hinting.general import DataClassType

LOG_LEVEL_MAP = {
    MessageLevel.Message_LEVEL_ERROR: logging.ERROR,
    MessageLevel.Message_LEVEL_WARNING: logging.WARN,
    MessageLevel.Message_LEVEL_INFO: logging.INFO,
}


def list_fields(node: DataClassType) -> Dict[str, Union[betterproto.Message, Iterable[betterproto.Message]]]:
    fields = dataclasses.fields(
        node
    )  # This type warining is OK, python are idiots and has no builtin type for dataclasses.
    output = {}
    for field in fields:
        field_value = getattr(node, field.name)
        if isinstance(field_value, Iterable) or (
            isinstance(field_value, betterproto.Message) and betterproto.serialized_on_wire(field_value)
        ):
            output[field.name] = field_value
    return output



import pretty_errors

pretty_errors.configure(filename_display=pretty_errors.FILENAME_COMPACT, display_arrow=True)


def disable_colored_errors():
    """Disables the colors in the error output, used for terminals that does not support coloring,
    such as powershell, cmder, and other physical terminals.
    """
    pretty_errors.mono()


def activate_verbose_errors():
    """Adds advance debug information to the error output, useful for debugging."""
    pretty_errors.configure(
        filename_display=pretty_errors.FILENAME_FULL,
        lines_before=5,
        lines_after=2,
        trace_lines_before=5,
        trace_lines_after=2,
        display_locals=True,
        display_trace_locals=True,
    )



from dataclasses import dataclass
from enum import Enum
from typing import List

from qm.grpc.results_analyser import GetJobErrorsResponseExecutionErrorSeverity
from qm.api.job_result_api import JobResultServiceApi


class ExecutionErrorSeverity(Enum):
    Warn = 0
    Error = 1


@dataclass(frozen=True)
class ExecutionError:
    error_code: int
    message: str
    severity: ExecutionErrorSeverity

    def __repr__(self):
        return f"{self.error_code}\t\t{self.severity.name}\t\t{self.message}"


class ExecutionReport:
    def __init__(self, job_id: str, service: JobResultServiceApi) -> None:
        self._job_id = job_id
        self._service = service
        self._errors = self._load_errors()

    def _load_errors(self) -> List[ExecutionError]:
        return [
            ExecutionError(
                error_code=item.error_code,
                message=item.message,
                severity=ExecutionReport._parse_error_severity(item.error_severity),
            )
            for item in self._service.get_job_errors(self._job_id)
        ]

    @staticmethod
    def _parse_error_severity(error_severity) -> ExecutionErrorSeverity:
        if error_severity == GetJobErrorsResponseExecutionErrorSeverity.WARNING:
            return ExecutionErrorSeverity.Warn
        elif error_severity == GetJobErrorsResponseExecutionErrorSeverity.ERROR:
            return ExecutionErrorSeverity.Error

    def has_errors(self) -> bool:
        """Returns: True if encountered a runtime error while executing the job."""
        return len(self._errors) > 0

    def errors(self) -> List[ExecutionError]:
        """Returns: list of all execution errors for this job"""
        return self._errors.copy()

    @property
    def _report_header(self) -> str:
        return (
            f"Execution report for job {self._job_id}\nErrors:\n"
            f"Please refer to section: "
            f"Error Indications and Error Reporting in documentation for additional information\n\n"
            "code\t\tseverity\tmessage"
        )

    def __repr__(self) -> str:
        if not self.has_errors():
            return f"Execution report for job {self._job_id}\nNo errors"

        errors_str = self._report_header
        for error in self._errors:
            errors_str += "\n" + str(error)
        return errors_str



import numpy as np


class SimulatorSamples:
    def __init__(self, controllers):
        for k, v in controllers.items():
            self.__setattr__(k, v)

    @staticmethod
    def from_np_array(arr: np.ndarray) -> "SimulatorSamples":
        controllers = dict()
        for col in arr.dtype.names:
            parts = col.split(":")
            controller = controllers.setdefault(parts[0], {"analog": dict(), "digital": dict()})
            controller[parts[1]][parts[2]] = arr[col]
        res = dict()
        for item in controllers.items():
            res[item[0]] = SimulatorControllerSamples(item[1]["analog"], item[1]["digital"])
        return SimulatorSamples(res)


class SimulatorControllerSamples(object):
    def __init__(self, analog, digital):
        self.analog = self._add_keys_for_first_con(analog)
        self.digital = self._add_keys_for_first_con(digital)
        self._analog_conns = analog
        self._digital_conns = digital

    @staticmethod
    def _add_keys_for_first_con(data):
        first_con_data = {k[2:]: v for (k, v) in data.items() if k.startswith("1-")}
        return {**data, **first_con_data}

    def plot(self, analog_ports=None, digital_ports=None):
        """Plots the simulated output of the OPX in the given ports.
        If no ports are given, all active ports are plotted.

        Args:
            analog_ports: Union[None, str, list[str]]
            digital_ports: Union[None, str, list[str]]
        """
        import matplotlib.pyplot as plt

        for port, samples in self._analog_conns.items():
            if analog_ports is None or port in analog_ports:
                plt.plot(samples, label=f"Analog {port}")
        for port, samples in self._digital_conns.items():
            if digital_ports is None or port in digital_ports:
                plt.plot(samples, label=f"Digital {port}")
        plt.xlabel("Time [ns]")
        plt.ylabel("Output")
        plt.legend()



from typing import Union, Optional

from qm.exceptions import QmQuaException
import qm.grpc.qua as _qua
from qm._loc import _get_loc

_ScalarExpressionType = _qua.QuaProgramAnyScalarExpression
_VariableRefType = _qua.QuaProgramVarRefExpression


def var(name: str) -> _VariableRefType:
    """A reference to a variable

    Args:
        name

    Returns:

    """
    exp = _qua.QuaProgramVarRefExpression(name=name, loc=_get_loc())
    return exp


def binary(left: _ScalarExpressionType, sop: str, right: _ScalarExpressionType) -> _ScalarExpressionType:
    """A binary operation

    Args:
        left
        sop
        right

    Returns:

    """
    if sop == "+":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.ADD
    elif sop == "-":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.SUB
    elif sop == ">":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.GT
    elif sop == "<":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.LT
    elif sop == "<=":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.LET
    elif sop == ">=":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.GET
    elif sop == "==":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.EQ
    elif sop == "*":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.MULT
    elif sop == "/":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.DIV
    elif sop == "|":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.OR
    elif sop == "&":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.AND
    elif sop == "^":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.XOR
    elif sop == "<<":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.SHL
    elif sop == ">>":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.SHR
    else:
        raise QmQuaException("Unsupported operator " + sop)

    exp = _qua.QuaProgramAnyScalarExpression(
        binary_operation=_qua.QuaProgramBinaryExpression(loc=_get_loc(), left=left, right=right, op=op)
    )
    return exp


def literal_int(value: int) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(
        literal=_qua.QuaProgramLiteralExpression(value=str(value), type=_qua.QuaProgramType.INT, loc=_get_loc())
    )
    return exp


def literal_bool(value: bool) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(
        literal=_qua.QuaProgramLiteralExpression(value=str(value), type=_qua.QuaProgramType.BOOL, loc=_get_loc())
    )
    return exp


def literal_real(value: float) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(
        literal=_qua.QuaProgramLiteralExpression(value=str(value), type=_qua.QuaProgramType.REAL, loc=_get_loc())
    )
    return exp


def io1() -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(variable=_qua.QuaProgramVarRefExpression(io_number=1, loc=_get_loc()))
    return exp


def io2() -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(variable=_qua.QuaProgramVarRefExpression(io_number=2, loc=_get_loc()))
    return exp


def array(
    value: Union[_qua.QuaProgramArrayVarRefExpression, _ScalarExpressionType],
    index_exp: Optional[_ScalarExpressionType],
) -> Union[_qua.QuaProgramArrayVarRefExpression, _ScalarExpressionType]:
    if index_exp is None:
        return value
    else:
        loc = _get_loc()
        value.loc = loc
        exp = _qua.QuaProgramAnyScalarExpression(
            array_cell=_qua.QuaProgramArrayCellRefExpression(array_var=value, index=index_exp, loc=loc)
        )
        return exp


def var_ref(value: str, index_exp: Optional[_ScalarExpressionType]) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression()

    loc = _get_loc()
    if index_exp is None:
        exp.variable = _qua.QuaProgramVarRefExpression(name=value, loc=loc)
    else:
        exp.array_cell = _qua.QuaProgramArrayCellRefExpression(
            array_var=_qua.QuaProgramArrayVarRefExpression(name=value, loc=loc),
            index=index_exp,
            loc=loc,
        )
    return exp


def lib_func(
    lib_name: str,
    func_name: str,
    *args: Union[_ScalarExpressionType, _qua.QuaProgramArrayVarRefExpression],
) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(
        lib_function=_qua.QuaProgramLibFunctionExpression(
            loc=_get_loc(), function_name=func_name, library_name=lib_name
        )
    )

    for arg in args:
        if isinstance(arg, _qua.QuaProgramArrayVarRefExpression):
            element = _qua.QuaProgramLibFunctionExpressionArgument(array=arg)
            exp.lib_function.arguments.append(element)
        else:
            element = _qua.QuaProgramLibFunctionExpressionArgument(scalar=arg)
            exp.lib_function.arguments.append(element)

    return exp



from typing import Dict, List

from typing_extensions import TypedDict


class WaveformOverrideType(TypedDict):
    samples: List[float]


class ExecutionOverridesType(TypedDict, total=False):
    waveforms: Dict[str, WaveformOverrideType]



import betterproto

from qm.utils import list_fields


class QuaNodeVisitor:
    def _default_enter(self, node):
        return True

    def _default_leave(self, node):
        pass

    def _default_visit(self, node):
        for _, field in list_fields(node).items():
            self.__visit(field)

    def __call(self, name, node, t):
        attr_name = f"{t}_{name}".replace(".", "_")
        attr = getattr(self, attr_name, None)
        if attr:
            ret = attr(node)
            if t == "enter":
                return ret
        elif t == "enter":
            return self._default_enter(node)
        elif t == "leave":
            self._default_leave(node)
        elif t == "visit":
            self._default_visit(node)
        else:
            raise Exception(f"unknown call type {t}. only 'enter', 'leave' or 'visit' are supported")

    def __call_enter(self, name, node):
        return self.__call(name, node, "enter")

    def __call_visit(self, name, node):
        return self.__call(name, node, "visit")

    def __call_leave(self, name, node):
        return self.__call(name, node, "leave")

    def visit(self, node):
        type_name = type(node).__name__
        type_module = type(node).__module__
        type_fullname = f"{type_module}.{type_name}"
        if type_fullname == "qm.program.program.Program":
            return self.visit(node._program)

        if isinstance(node, betterproto.Message):
            self.__visit(node)
        else:
            raise Exception(f"Failed to find descriptor on {node}")

    def __visit(self, node):
        if not isinstance(node, betterproto.Message):
            if isinstance(node, list):
                for n in node:
                    self.__visit(n)
            return
        self._enter_visit_leave(node)

    def _enter_visit_leave(self, node):
        type_fullname = f"{type(node).__module__}.{type(node).__name__}"
        if self.__call_enter(type_fullname, node):
            self.__call_visit(type_fullname, node)
        self.__call_leave(type_fullname, node)



from typing import List, Any, Mapping, Dict
import logging

import betterproto
import numpy as np
from betterproto.lib.google.protobuf import Empty
from dependency_injector.wiring import inject, Provide
from marshmallow import (
    Schema,
    fields,
    post_load,
    ValidationError,
    validates_schema,
    validate,
)
from marshmallow_polyfield import PolyField

from qm.api.models.capabilities import ServerCapabilities
from qm.containers.capabilities_container import CapabilitiesContainer
from qm.exceptions import ConfigValidationException
from qm.program._qua_config_to_pb import (
    octave_connection_to_pb,
    get_octave_loopbacks,
    OctaveConnectionAmbiguity,
    octave_params_to_pb,
)

from qm.program._validate_config_schema import (
    validate_output_smearing,
    validate_oscillator,
    validate_output_tof,
    validate_arbitrary_waveform,
    validate_used_inputs,
    validate_timetagging_parameters,
)
from qm.grpc import qua_config, octave_models
from qm.grpc.qua_config import QuaConfigAnalogOutputPortFilter, QuaConfig
from qm.type_hinting.config_types import DictQuaConfig

logger = logging.getLogger(__name__)


def validate_config_capabilities(pb_config, server_capabilities: ServerCapabilities):
    if not server_capabilities.supports_inverted_digital_output:
        for con_name, con in pb_config.v1_beta.controllers.items():
            for port_id, port in con.digital_outputs.items():
                if port.inverted:
                    raise ConfigValidationException(
                        f"Server does not support inverted digital output used in controller "
                        f"'{con_name}', port {port_id}"
                    )
    if not server_capabilities.supports_multiple_inputs_for_element:
        for el_name, el in pb_config.v1_beta.elements.items():
            if el is not None and el.multiple_inputs:
                raise ConfigValidationException(
                    f"Server does not support multiple inputs for elements used in '{el_name}'"
                )

    if not server_capabilities.supports_analog_delay:
        for con_name, con in pb_config.v1_beta.controllers.items():
            for port_id, port in con.analog_outputs.items():
                if port.delay != 0:
                    raise ConfigValidationException(
                        f"Server does not support analog delay used in controller " f"'{con_name}', port {port_id}"
                    )

    if not server_capabilities.supports_shared_oscillators:
        for el_name, el in pb_config.v1_beta.elements.items():
            if el is not None and el.named_oscillator:
                raise ConfigValidationException(
                    f"Server does not support shared oscillators for elements used in " f"'{el_name}'"
                )

    if not server_capabilities.supports_crosstalk:
        for con_name, con in pb_config.v1_beta.controllers.items():
            for port_id, port in con.analog_outputs.items():
                if len(port.crosstalk) > 0:
                    raise ConfigValidationException(
                        f"Server does not support channel weights used in controller " f"'{con_name}', port {port_id}"
                    )

    if not server_capabilities.supports_shared_ports:
        shared_ports_by_controller = {}
        for (con_name, con) in pb_config.v1_beta.controllers.items():
            shared_ports_by_type = {}
            analog_outputs = [port_id for port_id, port in con.analog_outputs.items() if port.shareable]
            analog_inputs = [port_id for port_id, port in con.analog_inputs.items() if port.shareable]
            digital_outputs = [port_id for port_id, port in con.digital_outputs.items() if port.shareable]
            digital_inputs = [port_id for port_id, port in con.digital_inputs.items() if port.shareable]
            if len(analog_outputs):
                shared_ports_by_type["analog_outputs"] = analog_outputs
            if len(analog_inputs):
                shared_ports_by_type["analog_inputs"] = analog_inputs
            if len(digital_outputs):
                shared_ports_by_type["digital_outputs"] = digital_outputs
            if len(digital_inputs):
                shared_ports_by_type["digital_inputs"] = digital_inputs
            if len(shared_ports_by_type):
                shared_ports_by_controller[con_name] = shared_ports_by_type
        if len(shared_ports_by_controller) > 0:
            error_message = "Server does not support shareable ports." + "\n".join(
                [
                    f"Controller: {con_name}\n{shared_ports_list}"
                    for con_name, shared_ports_list in shared_ports_by_controller.items()
                ]
            )
            raise ConfigValidationException(error_message)
    if not server_capabilities.supports_double_frequency:
        message_template = (
            "Server does not support float frequency. "
            "Element: {element_name}: {frequency_type}={float_value} "
            "will be casted to {int_value}."
        )
        for el_name, el in list(pb_config.v1_beta.elements.items()):
            if el.intermediate_frequency_double and el.intermediate_frequency_double != el.intermediate_frequency:
                logger.warning(
                    message_template.format(
                        element_name=el_name,
                        frequency_type="intermediate_frequency",
                        float_value=el.intermediate_frequency_double,
                        int_value=el.intermediate_frequency,
                    )
                )
            if (
                betterproto.serialized_on_wire(el.mix_inputs)
                and el.mix_inputs.lo_frequency_double
                and el.mix_inputs.lo_frequency != el.mix_inputs.lo_frequency_double
            ):
                logger.warning(
                    message_template.format(
                        element_name=el_name,
                        frequency_type="lo_frequency",
                        float_value=el.mix_inputs.lo_frequency_double,
                        int_value=el.mix_inputs.lo_frequency,
                    )
                )


def load_config(config: DictQuaConfig) -> QuaConfig:
    return QuaConfigSchema().load(config)


PortReferenceSchema = fields.Tuple(
    (fields.String(), fields.Int()),
    metadata={"description": "Controller port to use. Tuple of: ([str] controller name, [int] controller port)"},
)


class UnionField(fields.Field):
    """Field that deserializes multi-type input data to app-level objects."""

    def __init__(self, val_types: List[fields.Field], **kwargs):
        self.valid_types = val_types
        super().__init__(**kwargs)

    def _deserialize(self, value: Any, attr: str = None, data: Mapping[str, Any] = None, **kwargs):
        """
        _deserialize defines a custom Marshmallow Schema Field that takes in
        mutli-type input data to app-level objects.

        Parameters
        ----------
        value : {Any}
            The value to be deserialized.

        Keyword Parameters
        ----------
        attr : {str} [Optional]
            The attribute/key in data to be deserialized. (default: {None})
        data : {Optional[Mapping[str, Any]]}
            The raw input data passed to the Schema.load. (default: {None})

        Raises
        ----------
        ValidationError : Exception
            Raised when the validation fails on a field or schema.
        """
        errors = []
        # iterate through the types being passed into UnionField via val_types
        for field in self.valid_types:
            try:
                # inherit deserialize method from Fields class
                return field.deserialize(value, attr, data, **kwargs)
            # if error, add error message to error list
            except ValidationError as error:
                errors.append(error.messages)

        raise ValidationError(errors)


class AnalogOutputFilterDefSchema(Schema):
    feedforward = fields.List(
        fields.Float(),
        metadata={"description": "Feedforward taps for the analog output filter, range: [-1,1]. List of double"},
    )
    feedback = fields.List(
        fields.Float(),
        metadata={"description": "Feedback taps for the analog output filter, range: (-1,1). List of double"},
    )


class AnalogOutputPortDefSchema(Schema):
    offset = fields.Number(
        metadata={
            "description": "DC offset to the output, range: (-0.5, 0.5). "
            "Will be applied while quantum machine is open."
        }
    )
    filter = fields.Nested(AnalogOutputFilterDefSchema)
    delay = fields.Int(metadata={"description": "Output's delay, in units of ns."})
    crosstalk = fields.Dict(
        keys=fields.Int(), values=fields.Number(), metadata={"description": ""}
    )  # TODO: add description
    shareable = fields.Bool(
        dump_default=False,
        metadata={"description": "Whether the port is shareable with other QM instances"},
    )
    connectivity = fields.Tuple(
        (fields.String(), fields.String()),
        metadata={
            "description": "Connectivity (for now only for octave), Tuple of: ([str] Octave name, [str] port name)"
        },
    )

    class Meta:
        title = "Analog output port"
        description = "The specifications and properties of an analog output port of the controller."

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigAnalogOutputPortDec(
            offset=data["offset"],
            shareable=data.get("shareable", False),
        )

        if "delay" in data:
            delay = data.get("delay", 0)
            if delay < 0:
                raise ConfigValidationException(f"analog output delay cannot be a negative value, given value: {delay}")
            item.delay = delay

        if "filter" in data:
            filters = data["filter"]
            item.filter = QuaConfigAnalogOutputPortFilter(
                feedforward=filters.get("feedforward", []),
                feedback=filters.get("feedback", []),
            )

        if "crosstalk" in data:
            for k, v in data["crosstalk"].items():
                item.crosstalk[k] = v

        if "connectivity" in data:
            item.octave_connectivity = octave_connection_to_pb("input", port_data=data)

        return item


class AnalogInputPortDefSchema(Schema):
    offset = fields.Number(
        metadata={"description": "DC offset to the input, range: (-0.5, 0.5). Will be applied only when program runs."}
    )

    gain_db = fields.Int(
        strict=True,
        metadata={"description": "Gain of the pre-ADC amplifier, in dB. Accepts integers in the range: -12 to 20"},
    )

    shareable = fields.Bool(
        dump_default=False,
        metadata={"description": "Whether the port is shareable with other QM instances"},
    )

    connectivity = fields.Tuple(
        (fields.String(), fields.String()),
        metadata={
            "description": "Connectivity (for now only for octave), Tuple of: ([str] Octave name, [str] port name)"
        },
    )

    class Meta:
        title = "Analog input port"
        description = "The specifications and properties of an analog input port of the controller."

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigAnalogInputPortDec(
            offset=data.get("offset", 0.0),
            gain_db=data.get("gain_db"),
            shareable=data.get("shareable", False),
        )

        if "connectivity" in data:
            item.octave_connectivity = octave_connection_to_pb("output", port_data=data)

        return item


class DigitalOutputPortDefSchema(Schema):
    shareable = fields.Bool(
        dump_default=False,
        metadata={"description": "Whether the port is shareable with other QM instances"},
    )
    inverted = fields.Bool(
        dump_default=False,
        metadata={"description": "Whether the port is inverted. " "If True, the output will be inverted."},
    )

    connectivity = fields.Tuple(
        (fields.String(), fields.String()),
        metadata={
            "description": "Connectivity (for now only for octave), Tuple of: ([str] Octave name, [str] port name)"
        },
    )

    class Meta:
        title = "Digital port"
        description = "The specifications and properties of a digital output port of the controller."

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigDigitalOutputPortDec(
            shareable=data.get("shareable", False), inverted=data.get("inverted", False)
        )
        # TODO - once we decide how to add digital connectivity this can be added back
        # if "connectivity" in data:
        #     item.octave_connectivity = octave_connection_to_pb("digital_input", port_data=data)
        return item


class DigitalInputPortDefSchema(Schema):
    deadtime = fields.Int(metadata={"description": "The minimal time between pulses, in ns."})
    polarity = fields.String(
        metadata={"description": "The Detection edge - Whether to trigger in the rising or falling edge of the pulse"},
        validate=validate.OneOf(["RISING", "FALLING"]),
    )
    threshold = fields.Number(metadata={"description": "The minimum voltage to trigger when a pulse arrives"})
    shareable = fields.Bool(
        dump_default=False,
        metadata={"description": "Whether the port is shareable with other QM instances"},
    )
    connectivity = fields.Tuple(
        (fields.String(), fields.String()),
        metadata={
            "description": "Connectivity (for now only for octave), Tuple of: ([str] Octave name, [str] port name)"
        },
    )

    class Meta:
        title = "Digital input port"
        description = "The specifications and properties of a digital input " "port of the controller."

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigDigitalInputPortDec()
        item.deadtime = data["deadtime"]
        if data["polarity"].upper() == "RISING":
            item.polarity = qua_config.QuaConfigDigitalInputPortDecPolarity.RISING
        elif data["polarity"].upper() == "FALLING":
            item.polarity = qua_config.QuaConfigDigitalInputPortDecPolarity.FALLING

        item.threshold = data["threshold"]
        if "shareable" in data:
            item.shareable = data["shareable"]
        # TODO - once we decide how to add digital connectivity this can be added back
        # if "connectivity" in data:
        #     item.octave_connectivity = octave_connection_to_pb("digital_output", port_data=data)
        return item


class OctaveSchema(Schema):
    loopbacks = fields.List(
        fields.Tuple([fields.Tuple([fields.String, fields.String]), fields.String]),
        metadata={
            "description": "List of loopbacks that connected to this octave, Each loopback is "
            "in the form of ((octave_name, octave_port), target_port)"
        },
    )

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = octave_models.OctaveConfig()
        item.loopbacks = get_octave_loopbacks(data["loopbacks"])
        return item


class ControllerSchema(Schema):
    type = fields.Constant("opx1")

    analog_outputs = fields.Dict(
        fields.Int(),
        fields.Nested(AnalogOutputPortDefSchema),
        metadata={"description": "The analog output ports and their properties."},
    )
    analog_inputs = fields.Dict(
        fields.Int(),
        fields.Nested(AnalogInputPortDefSchema),
        metadata={"description": "The analog input ports and their properties."},
    )
    digital_outputs = fields.Dict(
        fields.Int(),
        fields.Nested(DigitalOutputPortDefSchema),
        metadata={"description": "The digital output ports and their properties."},
    )
    digital_inputs = fields.Dict(
        fields.Int(),
        fields.Nested(DigitalInputPortDefSchema),
        metadata={"description": "The digital inputs ports and their properties."},
    )
    connectivity = fields.String(
        metadata={
            "description": "Here you can specify the connection of the OPX, for now, you can put here the name "
            "of the octave and it will set the default connection between OPX and Octave."
        }
    )

    class Meta:
        title = "controller"
        description = "The specification of a single controller and its properties."

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigControllerDec()
        item.type = data["type"]

        if "analog_outputs" in data:
            for k, v in data["analog_outputs"].items():
                item.analog_outputs[k] = v
                if "connectivity" in data:
                    if betterproto.serialized_on_wire(v.octave_connectivity):
                        raise OctaveConnectionAmbiguity
                    item.analog_outputs[k].octave_connectivity = octave_connection_to_pb(
                        "input", device_name=data["connectivity"], port_num=k
                    )

        if "analog_inputs" in data:
            for k, v in data["analog_inputs"].items():
                item.analog_inputs[k] = v
                if "connectivity" in data:
                    if betterproto.serialized_on_wire(v.octave_connectivity):
                        raise OctaveConnectionAmbiguity
                    item.analog_inputs[k].octave_connectivity = octave_connection_to_pb(
                        "output", device_name=data["connectivity"], port_num=k
                    )

        if "digital_outputs" in data:
            for k, v in data["digital_outputs"].items():
                item.digital_outputs[k] = v

        if "digital_inputs" in data:
            for k, v in data["digital_inputs"].items():
                item.digital_inputs[k] = v

        return item


class DigitalInputSchema(Schema):
    delay = fields.Int(
        metadata={
            "description": "The delay to apply to the digital pulses. In ns. "
            "An intrinsic negative delay of 136 ns exists by default"
        }
    )
    buffer = fields.Int(
        metadata={
            "description": "Digital pulses played to this element will be convolved with a digital "
            "pulse of value 1 with this length [ns]"
        }
    )
    port = PortReferenceSchema

    class Meta:
        title = "Digital input"
        description = "The specification of the digital input of an element"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigDigitalInputPortReference()
        item.delay = data["delay"]
        item.buffer = data["buffer"]
        if "port" in data:
            item.port.controller = data["port"][0]
            item.port.number = data["port"][1]
        return item


class IntegrationWeightSchema(Schema):
    cosine = UnionField(
        [
            fields.List(fields.Tuple([fields.Float(), fields.Int()])),
            fields.List(fields.Float()),
        ],
        metadata={
            "description": "The integration weights for the cosine. Given as a list of tuples, "
            "each tuple in the format of: ([double] weight, [int] duration). "
            "weight range: [-2048, 2048] in steps of 2**-15. duration is in ns "
            "and must be a multiple of 4."
        },
    )
    sine = UnionField(
        [
            fields.List(fields.Tuple([fields.Float(), fields.Int()])),
            fields.List(fields.Float()),
        ],
        metadata={
            "description": "The integration weights for the sine. Given as a list of tuples, "
            "each tuple in the format of: ([double] weight, [int] duration). "
            "weight range: [-2048, 2048] in steps of 2**-15. duration is in ns "
            "and must be a multiple of 4."
        },
    )

    class Meta:
        title = "Integration weights"
        description = "The specification of measurements' integration weights."

    @staticmethod
    def build_iw_sample(data):
        if len(data) > 0 and not isinstance(data[0], tuple):
            data = np.round(2**-15 * np.round(np.array(data) / 2**-15), 20)
            new_data = []
            for s in data:
                if len(new_data) == 0 or new_data[-1][0] != s:
                    new_data.append((s, 4))
                else:
                    new_data[-1] = (new_data[-1][0], new_data[-1][1] + 4)
            data = new_data
        res = []
        for s in data:
            sample = qua_config.QuaConfigIntegrationWeightSample(value=s[0], length=s[1])
            res.append(sample)
        return res

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigIntegrationWeightDec()
        if "cosine" in data:
            item.cosine.extend(self.build_iw_sample(data["cosine"]))
        if "sine" in data:
            item.sine.extend(self.build_iw_sample(data["sine"]))
        return item


class WaveFormSchema(Schema):
    pass


class ConstantWaveFormSchema(WaveFormSchema):
    type = fields.String(metadata={"description": '"constant"'})
    sample = fields.Float(metadata={"description": "Waveform amplitude, range: (-0.5, 0.5)"})

    class Meta:
        title = "Constant waveform"
        description = "A waveform with a constant amplitude"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigWaveformDec(constant=qua_config.QuaConfigConstantWaveformDec(sample=data["sample"]))
        return item


class ArbitraryWaveFormSchema(WaveFormSchema):
    type = fields.String(metadata={"description": '"arbitrary"'})
    samples = fields.List(
        fields.Float(),
        metadata={"description": "list of values of an arbitrary waveforms, range: (-0.5, 0.5)"},
    )
    max_allowed_error = fields.Float(metadata={"description": '"Maximum allowed error for automatic compression"'})
    sampling_rate = fields.Number(
        metadata={
            "description": "Sampling rate to use in units of S/s (samples per second). "
            "Default is 1e9. Cannot be set when is_overridable=True"
        }
    )
    is_overridable = fields.Bool(
        dump_default=False,
        metadata={
            "description": "Allows overriding the waveform after compilation. "
            "Cannot use with the non-default sampling_rate"
        },
    )

    class Meta:
        title = "arbitrary waveform"
        description = "The modulating envelope of an arbitrary waveform"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        is_overridable = data.get("is_overridable", False)
        item = qua_config.QuaConfigWaveformDec(
            arbitrary=qua_config.QuaConfigArbitraryWaveformDec(samples=data["samples"], is_overridable=is_overridable)
        )
        max_allowed_error_key = "max_allowed_error"
        sampling_rate_key = "sampling_rate"
        has_max_allowed_error = max_allowed_error_key in data
        has_sampling_rate = sampling_rate_key in data
        validate_arbitrary_waveform(is_overridable, has_max_allowed_error, has_sampling_rate)
        if has_max_allowed_error:
            item.arbitrary.max_allowed_error = data[max_allowed_error_key]
        elif has_sampling_rate:
            item.arbitrary.sampling_rate = data[sampling_rate_key]
        elif not is_overridable:
            item.arbitrary.max_allowed_error = 1e-4
        return item


def _waveform_schema_deserialization_disambiguation(object_dict, data):
    type_to_schema = {
        "constant": ConstantWaveFormSchema,
        "arbitrary": ArbitraryWaveFormSchema,
    }
    try:
        return type_to_schema[object_dict["type"]]()
    except KeyError:
        pass

    raise TypeError("Could not detect type. Did not have a base or a length. Are you sure this is a shape?")


_waveform_poly_field = PolyField(
    deserialization_schema_selector=_waveform_schema_deserialization_disambiguation,
    required=True,
)


class DigitalWaveFormSchema(Schema):
    samples = fields.List(
        fields.Tuple([fields.Int(), fields.Int()]),
        metadata={
            "description": "The digital waveform. Given as a list of tuples, each tuple in the format of: "
            "([int] state, [int] duration). state is either 0 or 1 indicating whether the "
            "digital output is off or on. duration is in ns. If the duration is 0, "
            "it will be played until the reminder of the analog pulse"
        },
    )

    class Meta:
        title = "Digital waveform"
        description = "The samples of a digital waveform"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigDigitalWaveformDec()
        for sample in data["samples"]:
            item.samples.append(qua_config.QuaConfigDigitalWaveformSample(value=bool(sample[0]), length=int(sample[1])))
        return item


class MixerSchema(Schema):
    intermediate_frequency = fields.Float(
        metadata={"description": "The intermediate frequency associated with the correction matrix"}
    )
    lo_frequency = fields.Float(metadata={"description": "The LO frequency associated with the correction matrix"})
    correction = fields.Tuple(
        (fields.Number(), fields.Number(), fields.Number(), fields.Number()),
        metadata={
            "description": "A 2x2 matrix entered as a 4 elements list specifying the "
            "correction matrix. Each element is a double in the range of (-2,2)"
        },
    )

    class Meta:
        title = "Mixer"
        description = (
            "The specification of the correction matrix for an IQ mixer. "
            "This is a list of correction matrices for each LO and IF frequencies."
        )

    @post_load(pass_many=False)
    @inject
    def build(
        self,
        data,
        capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
        **kwargs,
    ):
        item = qua_config.QuaConfigCorrectionEntry(correction=qua_config.QuaConfigMatrix(*data["correction"]))

        if "lo_frequency" in data:
            if capabilities.supports_double_frequency:
                item.lo_frequency_double = float(data["lo_frequency"])
            else:
                item.lo_frequency = int(data["lo_frequency"])  # backwards compatibility

        if "intermediate_frequency" in data:
            if capabilities.supports_double_frequency:
                item.frequency_double = abs(float(data["intermediate_frequency"]))
            else:
                item.frequency = abs(int(data["intermediate_frequency"]))  # backwards compatibility

            item.frequency_negative = data["intermediate_frequency"] < 0

        item.correction = qua_config.QuaConfigMatrix(
            data["correction"][0],
            data["correction"][1],
            data["correction"][2],
            data["correction"][3],
        )
        return item


class PulseSchema(Schema):
    operation = fields.String(
        metadata={"description": "The type of operation. Possible values: 'control', 'measurement'"}
    )
    length = fields.Int(
        metadata={"description": "The length of pulse [ns]. Possible values: 16 to 2^31-1 in steps of 4"}
    )
    waveforms = fields.Dict(
        fields.String(),
        fields.String(metadata={"description": "The name of analog waveform to be played."}),
        metadata={
            "description": "The specification of the analog waveform to be played. "
            "If the associated element has a single input, then the key is 'single'. "
            "If the associated element has 'mixInputs', then the keys are 'I' and 'Q'."
        },
    )
    digital_marker = fields.String(
        metadata={"description": "The name of the digital waveform to be played with this pulse."}
    )
    integration_weights = fields.Dict(
        fields.String(),
        fields.String(
            metadata={
                "description": "The name of the integration weights as it appears under the"
                ' "integration_weigths" entry in the configuration dict.'
            }
        ),
        metadata={"description": "The name of the integration weight to be used in the program."},
    )

    class Meta:
        title = "pulse"
        description = "The specification and properties of a single pulse and to the measurement associated with it."

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigPulseDec()
        item.length = data["length"]
        if data["operation"] == "measurement":
            item.operation = qua_config.QuaConfigPulseDecOperation.MEASUREMENT
        elif data["operation"] == "control":
            item.operation = qua_config.QuaConfigPulseDecOperation.CONTROL
        if "integration_weights" in data:
            for k, v in data["integration_weights"].items():
                item.integration_weights[k] = v
        if "waveforms" in data:
            for k, v in data["waveforms"].items():
                item.waveforms[k] = v
        if "digital_marker" in data:
            item.digital_marker = data["digital_marker"]
        return item


class SingleInputSchema(Schema):
    port = PortReferenceSchema

    class Meta:
        title = "Single input"
        description = "The specification of the input of an element which has a single input port"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        controller, number = data["port"]
        item = qua_config.QuaConfigSingleInput(
            port=qua_config.QuaConfigDacPortReference(controller=controller, number=number)
        )
        return item


class HoldOffsetSchema(Schema):
    duration = fields.Int(metadata={"description": """The ramp to zero duration, in ns"""})

    class Meta:
        title = "Hold offset"
        description = "When defined, makes the element sticky"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigHoldOffset()
        item.duration = data["duration"]
        return item


class MixInputOctaveParamsSchema(Schema):
    rf_output_port = fields.Tuple([fields.String(), fields.String()])
    lo_source = fields.String()
    output_switch_state = fields.String()
    output_gain = fields.Int()
    downconversion_lo_source = fields.String()
    downconversion_lo_frequency = fields.Float()
    rf_input_port = fields.Tuple([fields.String(), fields.String()])

    @post_load()
    def build(
        self,
        data,
        **kwargs,
    ):
        return octave_params_to_pb(data)


class StickySchema(Schema):
    analog = fields.Boolean(metadata={"description": """the analog flag must be a True of False"""})
    digital = fields.Boolean(metadata={"description": """the digital flag must be a True of False"""})
    duration = fields.Int(metadata={"description": """The ramp to zero duration, in ns"""})

    class Meta:
        title = "Sticky"
        description = "When defined, makes the element sticky"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigSticky()
        item.duration = data.get("duration")
        item.analog = data.get("analog")
        if "digital" in data:
            item.digital = data.get("digital")
        return item


class MixInputSchema(Schema):
    I = PortReferenceSchema
    Q = PortReferenceSchema
    mixer = fields.String(
        metadata={
            "description": "The mixer used to drive the input of the element, "
            "taken from the names in mixers entry in the main configuration."
        }
    )
    lo_frequency = fields.Float(
        metadata={"description": "The frequency of the local oscillator which drives the mixer."}
    )
    octave_params = fields.Nested(MixInputOctaveParamsSchema)

    class Meta:
        title = "Mixer input"
        description = "The specification of the input of an element which is driven by an IQ mixer"

    @post_load(pass_many=False)
    @inject
    def build(
        self,
        data,
        capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
        **kwargs,
    ):
        lo_frequency = data.get("lo_frequency", 0)
        if capabilities.supports_double_frequency:
            frequency = {"lo_frequency_double": float(lo_frequency)}
        else:
            frequency = {"lo_frequency": int(lo_frequency)}

        item = qua_config.QuaConfigMixInputs(
            i=qua_config.QuaConfigDacPortReference(controller=data["I"][0], number=data["I"][1]),
            q=qua_config.QuaConfigDacPortReference(controller=data["Q"][0], number=data["Q"][1]),
            mixer=data.get("mixer", ""),
            **frequency,
        )
        if "octave_params" in data:
            item.octave_params = data["octave_params"]
        return item


class SingleInputCollectionSchema(Schema):
    inputs = fields.Dict(
        keys=fields.String(),
        values=PortReferenceSchema,
        metadata={"description": "A collection of multiple single inputs to the port"},
    )

    class Meta:
        title = "Single input collection"
        description = "Defines a set of single inputs which can be switched during play statements"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigSingleInputCollection()
        for name, (controller, number) in data["inputs"].items():
            item.inputs[name] = qua_config.QuaConfigDacPortReference(controller=controller, number=number)
        return item


class MultipleInputsSchema(Schema):
    inputs = fields.Dict(
        keys=fields.String(),
        values=PortReferenceSchema,
        metadata={"description": "A collection of multiple single inputs to the port"},
    )

    class Meta:
        title = "Multiple inputs"
        description = "Defines a set of single inputs which are all played at once"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        item = qua_config.QuaConfigMultipleInputs()
        for name, (controller, number) in data["inputs"].items():
            item.inputs[name] = qua_config.QuaConfigDacPortReference(controller=controller, number=number)
        return item


class OscillatorSchema(Schema):
    intermediate_frequency = fields.Float(
        metadata={"description": "The frequency of this oscillator [Hz]."},
        allow_none=True,
    )
    mixer = fields.String(
        metadata={
            "description": "The mixer used to drive the input of the oscillator, "
            "taken from the names in mixers entry in the main configuration"
        }
    )
    lo_frequency = fields.Float(
        metadata={"description": "The frequency of the local oscillator which drives the mixer [Hz]."}
    )

    @post_load(pass_many=False)
    @inject
    def build(
        self,
        data,
        capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
        **kwargs,
    ):
        osc = qua_config.QuaConfigOscillator()
        if "intermediate_frequency" in data and data["intermediate_frequency"] is not None:
            if capabilities.supports_double_frequency:
                osc.intermediate_frequency_double = data["intermediate_frequency"]
            else:
                osc.intermediate_frequency = int(data["intermediate_frequency"])
        if "mixer" in data and data["mixer"] is not None:
            osc.mixer.mixer = data["mixer"]
            if capabilities.supports_double_frequency:
                osc.mixer.lo_frequency_double = data.get("lo_frequency", 0.0)
            else:
                osc.mixer.lo_frequency = int(data.get("lo_frequency", 0))

        return osc


class ElementSchema(Schema):
    intermediate_frequency = fields.Float(
        metadata={"description": "The frequency at which the controller modulates the output to this element [Hz]."},
        allow_none=True,
    )
    oscillator = fields.String(
        metadata={
            "description": "The oscillator which is used by the controller to modulates the "
            "output to this element [Hz]. Can be used to share oscillators between elements"
        },
        allow_none=True,
    )

    measurement_qe = fields.String(metadata={"description": "not implemented"})
    operations = fields.Dict(
        keys=fields.String(),
        values=fields.String(
            metadata={
                "description": 'The name of the pulse as it appears under the "pulses" entry in the configuration dict'
            }
        ),
        metadata={"description": "A collection of all pulse names to be used in play and measure commands"},
    )
    singleInput = fields.Nested(SingleInputSchema)
    mixInputs = fields.Nested(MixInputSchema)
    singleInputCollection = fields.Nested(SingleInputCollectionSchema)
    multipleInputs = fields.Nested(MultipleInputsSchema)
    time_of_flight = fields.Int(
        metadata={
            "description": """The delay time, in ns, from the start of pulse until it reaches 
            the controller. Needs to be calibrated by looking at the raw ADC data. 
            Needs to be a multiple of 4 and the minimal value is 24. """
        }
    )
    smearing = fields.Int(
        metadata={
            "description": """Padding time, in ns, to add to both the start and end of the raw 
            ADC data window during a measure command."""
        }
    )
    outputs = fields.Dict(
        keys=fields.String(),
        values=PortReferenceSchema,
        metadata={"description": "The output ports of the element."},
    )
    digitalInputs = fields.Dict(keys=fields.String(), values=fields.Nested(DigitalInputSchema))
    digitalOutputs = fields.Dict(keys=fields.String(), values=PortReferenceSchema)
    outputPulseParameters = fields.Dict(metadata={"description": "Pulse parameters for Time-Tagging"})

    hold_offset = fields.Nested(HoldOffsetSchema)

    sticky = fields.Nested(StickySchema)

    thread = fields.String(metadata={"description": "QE thread"})

    class Meta:
        title = "Element"
        description = "The specifications, parameters and connections of a single element."

    @post_load(pass_many=False)
    @inject
    def build(
        self,
        data,
        capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
        **kwargs,
    ):
        el = qua_config.QuaConfigElementDec()
        if "intermediate_frequency" in data and data["intermediate_frequency"] is not None:
            if capabilities.supports_double_frequency:
                el.intermediate_frequency_double = abs(data["intermediate_frequency"])
                el.intermediate_frequency_oscillator_double = data["intermediate_frequency"]
            else:
                el.intermediate_frequency = abs(int(data["intermediate_frequency"]))
                el.intermediate_frequency_oscillator = int(data["intermediate_frequency"])

            el.intermediate_frequency_negative = data["intermediate_frequency"] < 0
        elif "oscillator" in data and data["oscillator"] is not None:
            el.named_oscillator = data["oscillator"]
        else:
            el.no_oscillator = Empty()

        # validate we have only 1 set of input defined
        validate_used_inputs(data)

        if "singleInput" in data:
            el.single_input = data["singleInput"]
        if "mixInputs" in data:
            el.mix_inputs = data["mixInputs"]
        if "singleInputCollection" in data:
            el.single_input_collection = data["singleInputCollection"]
        if "multipleInputs" in data:
            el.multiple_inputs = data["multipleInputs"]
        if "measurement_qe" in data:
            el.measurement_qe = data["measurement_qe"]
        if "time_of_flight" in data:
            el.time_of_flight = data["time_of_flight"]
        if "smearing" in data:
            el.smearing = data["smearing"]
        if "operations" in data:
            for k, v in data["operations"].items():
                el.operations[k] = v
        if "inputs" in data:
            el.inputs = _build_port(data["inputs"])
        if "outputs" in data:
            el.outputs = _build_port(data["outputs"])
        if "digitalInputs" in data:
            for k, v in data["digitalInputs"].items():
                el.digital_inputs[k] = v
        if "digitalOutputs" in data:
            for k, v in data["digitalOutputs"].items():
                el.digital_outputs[k] = qua_config.QuaConfigDigitalOutputPortReference(
                    port=qua_config.QuaConfigPortReference(controller=v[0], number=v[1])
                )
        if "outputPulseParameters" in data:
            pulse_parameters = data["outputPulseParameters"]
            el.output_pulse_parameters.signal_threshold = pulse_parameters["signalThreshold"]

            signal_polarity = pulse_parameters["signalPolarity"].upper()
            if signal_polarity == "ABOVE" or signal_polarity == "ASCENDING":
                el.output_pulse_parameters.signal_polarity = qua_config.QuaConfigOutputPulseParametersPolarity.ASCENDING
            elif signal_polarity == "BELOW" or signal_polarity == "DESCENDING":
                el.output_pulse_parameters.signal_polarity = (
                    qua_config.QuaConfigOutputPulseParametersPolarity.DESCENDING
                )

            if "derivativeThreshold" in pulse_parameters:
                el.output_pulse_parameters.derivative_threshold = pulse_parameters["derivativeThreshold"]

                polarity = pulse_parameters["derivativePolarity"].upper()
                if polarity == "ABOVE" or polarity == "ASCENDING":
                    el.output_pulse_parameters.derivative_polarity = (
                        qua_config.QuaConfigOutputPulseParametersPolarity.ASCENDING
                    )
                elif polarity == "BELOW" or polarity == "DESCENDING":
                    el.output_pulse_parameters.derivative_polarity = (
                        qua_config.QuaConfigOutputPulseParametersPolarity.DESCENDING
                    )
        if "sticky" in data:
            if capabilities.supports_sticky_elements:
                el.sticky = data["sticky"]
            else:
                if data["sticky"].digital:
                    raise ConfigValidationException(
                        f"Server does not support digital sticky used in element " f"'{el}'"
                    )
                el.hold_offset = qua_config.QuaConfigHoldOffset(duration=data["sticky"].duration)

        elif "hold_offset" in data:
            if capabilities.supports_sticky_elements:
                el.sticky = qua_config.QuaConfigSticky(
                    analog=True, digital=False, duration=data["hold_offset"].duration
                )
            else:
                el.hold_offset = data["hold_offset"]

        if "thread" in data:
            el.thread.thread_name = data["thread"]
        return el

    @validates_schema
    def validate_timetagging_parameters(self, data, **kwargs):
        validate_timetagging_parameters(data)

    @validates_schema
    def validate_output_tof(self, data, **kwargs):
        validate_output_tof(data)

    @validates_schema
    def validate_output_smearing(self, data, **kwargs):
        validate_output_smearing(data)

    @validates_schema
    def validate_oscillator(self, data, **kwargs):
        validate_oscillator(data)


def _build_port(data) -> Dict[str, qua_config.QuaConfigAdcPortReference]:
    outputs = {}
    if data is not None:
        for k, (controller, number) in data.items():
            outputs[k] = qua_config.QuaConfigAdcPortReference(controller=controller, number=number)
    return outputs


class QuaConfigSchema(Schema):
    version = fields.Int(metadata={"description": "Config version."})
    oscillators = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(OscillatorSchema),
        metadata={
            "description": """The oscillators used to drive the elements. 
        Can be used to share oscillators between elements"""
        },
    )

    elements = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(ElementSchema),
        metadata={
            "description": """The elements. Each element represents and
         describes a controlled entity which is connected to the ports of the 
         controller."""
        },
    )

    controllers = fields.Dict(
        fields.String(),
        fields.Nested(ControllerSchema),
        metadata={"description": """The controllers. """},
    )

    octaves = fields.Dict(
        fields.String(),
        fields.Nested(OctaveSchema),
        metadata={"description": "The octaves that are in the system, with their interconnected loopbacks."},
    )

    integration_weights = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(IntegrationWeightSchema),
        metadata={
            "description": """The integration weight vectors used in the integration 
        and demodulation of data returning from a element."""
        },
    )

    waveforms = fields.Dict(
        keys=fields.String(),
        values=_waveform_poly_field,
        metadata={
            "description": """The analog waveforms sent to an element when a pulse is 
        played."""
        },
    )
    digital_waveforms = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(DigitalWaveFormSchema),
        metadata={
            "description": """The digital waveforms sent to an element when a pulse is 
        played."""
        },
    )
    pulses = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(PulseSchema),
        metadata={"description": """The pulses to be played to the elements. """},
    )
    mixers = fields.Dict(
        keys=fields.String(),
        values=fields.List(fields.Nested(MixerSchema)),
        metadata={
            "description": """The IQ mixer calibration properties, used to post-shape the pulse
         to compensate for imperfections in the mixers used for up-converting the 
         analog waveforms."""
        },
    )

    class Meta:
        title = "QUA Config"
        description = "QUA program config root object"

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        config_wrapper = qua_config.QuaConfig()
        config = qua_config.QuaConfigQuaConfigV1()
        version = data["version"]
        if str(version) != "1":
            raise RuntimeError("Version must be set to 1 (was set to " + str(version) + ")")
        if "elements" in data:
            for el_name, el in data["elements"].items():
                config.elements[el_name] = el
        if "oscillators" in data:
            for osc_name, osc in data["oscillators"].items():
                config.oscillators[osc_name] = osc
        if "controllers" in data:
            for k, v in data["controllers"].items():
                config.controllers[k] = v
        if "octaves" in data:
            for k, v in data["octaves"].items():
                config.octaves[k] = v
        if "integration_weights" in data:
            for k, v in data["integration_weights"].items():
                config.integration_weights[k] = v
        if "waveforms" in data:
            for k, v in data["waveforms"].items():
                config.waveforms[k] = v
        if "digital_waveforms" in data:
            for k, v in data["digital_waveforms"].items():
                config.digital_waveforms[k] = v
        if "mixers" in data:
            for k, v in data["mixers"].items():
                config.mixers[k] = qua_config.QuaConfigMixerDec(correction=v)
        if "pulses" in data:
            for k, v in data["pulses"].items():
                config.pulses[k] = v
        config_wrapper.v1_beta = config
        return config_wrapper
